
//
// bubl.h
//
// This is the low level description of the object structure
// 
//
// $Id: bubl.h,v 1.5 2000/12/30 10:39:46 dru Exp $
//
// $Log: bubl.h,v $
// Revision 1.5  2000/12/30 10:39:46  dru
// This is the checkin that has working blocks!
// Plus tons of other fixes!
//
// Revision 1.4  2000/08/27 19:49:27  dru
// This is the result of a major change in the code to handle
// the new reference size, and the changes in the way the
// code handles activations.
//
// Revision 1.3  2000/04/26 01:50:04  dru
// This is a large set of changes that allowed aqua
// to do things properly. The most important of
// those was module importing. Lots of things
// were fixed (like the Lobby) and lots of intrinsics
// were added.
//
// Revision 1.2  1999/10/18 07:33:18  dru
// Lots and lots of changes:
// cyclic sd2bubl/bubl2sd
// fixes to doAquaInit, activateSlot, etc.
// fixes to runtime all over the place. This is a working runtime.
// This is the 'back on track' check in.
//
// Revision 1.1  1999/03/13 01:06:06  dnelson
// It wasn't in before
//
//
//
//
//


//
// This code is a collection of defines, structures, and 
// macros for working with protos on a low-level
//

#include "aqmalloc.h"

#ifndef H_BUBL
#define H_BUBL


//typedef unsigned int  uint;
typedef unsigned char uchar;


// These are the types encoded into object references

#define PROTO_REF_PROTO    0x00000000
#define PROTO_REF_INT      0x00000001
#define PROTO_REF_BLOB     0x00000002
#define PROTO_REF_ATOM     0x00000003
#define PROTO_REF_CODE     0x00000004
#define PROTO_REF_CONTEXT  0x00000005
#define PROTO_REF_BLOCK    0x00000006
#define PROTO_REF_FLOAT    0x00000007
#define PROTO_REF_MASK     (~PROTO_REF_FLOAT)
#define PROTO_REF_TYPE     (PROTO_REF_FLOAT)

// Get the type
#define PROTO_REFTYPE(p)      ( ( (uint32) (p) & PROTO_REF_TYPE ) )
// Get the actual ptr
#define PROTO_REFVALUE(p)     ( (Proto) ( (uint32) (p) & PROTO_REF_MASK ) )

// This is an object which is contained in the reference
// oddballs
#define PROTO_REFTYPE_REF(p)    ( (PROTO_REFTYPE(p) & 0x01) )

// This is an object which is contained in the heap
#define PROTO_REFTYPE_HEAP(p)   ( ((PROTO_REFTYPE(p) & 0x01) == 0) )
#define PROTO_REFTYPE_PROTO(p)  ( (((PROTO_REFTYPE(p) & 0x01) == 0) && (PROTO_REFTYPE(p) != 2)) )




// These are fields related to sections of an allocated proto
// on the heap.

struct _protoEntry 
{
    uint32 keyActionIndex;	// Lower 8 bits are the index, the next 3 bits indicate the action, the upper 21 are the key (an atom)
};

typedef struct _protoEntry * ProtoSlot;


    
#define PROTO_SIZEFIELD   0x7FFFFFF0

// Garbage collectable / opposite is manual collection 
#define PROTO_FLAG_PREVINUSE    0x00000001
#define PROTO_FLAG_GC           0x00000002
// Is a proto  / opposite is a blob (this only applies if GC is set)
#define PROTO_FLAG_PROTO        0x00000004
#define PROTO_FLAG_BLOB         0x00000000

#define PROTO_FLAG_FORWARDER    0x00000008
#define PROTO_FLAG_MARKED       0x80000000

// See if a proto is a forwarder
#define PROTO_ISFORWARDER(p)     ( (PROTO_REFVALUE(p))->header & PROTO_FLAG_FORWARDER )
#define PROTO_FORWARDVALUE(p)    ( (Proto) ((PROTO_REFVALUE(p))->count) )

// This is the special case of the proto's key in a slot
// Since we use the lower 3 bits for the action code, the
// Key must have these masked out. So, when atom pointers
// are returned by that module, they are acutally left shifted
// by 1 
#define PROTO_SLOT_KEYMASK     (0xFFFFF800)
#define PROTO_SLOT_ACTIONMASK  (0x700)
#define PROTO_SLOT_KEYACTIONMASK     (PROTO_SLOT_KEYMASK | PROTO_SLOT_ACTIONMASK)
#define PROTO_SLOT_ACTIONMASK  (0x700)
#define PROTO_SLOT_INDEXMASK   (0xFF)

// Notice that we don't shift down the key, this is because it's normal form
// is left shifted 11 binary digits
#define slotKey(p)    ( ((p)->keyActionIndex & PROTO_SLOT_KEYMASK) | PROTO_REF_ATOM  ) 
#define slotAction(p) (((p)->keyActionIndex & PROTO_SLOT_ACTIONMASK) >> 8  )
#define slotIndex(p)  (((p)->keyActionIndex & PROTO_SLOT_INDEXMASK)  >> 0  )

// Again, notice that we don't shift down the key, this is because it's normal form
// is left shifted 11 binary digits, plus we remove the PROTO_REF_MASK bits
// with the argument before it is OR'd in
#define slotSetKey(p,k)     ((p)->keyActionIndex = ((p)->keyActionIndex & ~PROTO_SLOT_KEYMASK)    |  ((k) & PROTO_SLOT_KEYMASK) )
#define slotSetAction(p,a)  ((p)->keyActionIndex = ((p)->keyActionIndex & ~PROTO_SLOT_ACTIONMASK) | ((a) << 8)  )
#define slotSetIndex(p,i)   ((p)->keyActionIndex = ((p)->keyActionIndex & ~PROTO_SLOT_INDEXMASK)  |  (i)        )



// Proto Size Field 

#define PROTO_COUNT_DATAMASK     (0xFF)
#define PROTO_COUNT_SLOTMASK     (0xFF00)
// deprecated because we have the new reference types
#define PROTO_COUNT_TYPEISCODE   0x80000000


#define protoDataCount(p)    (((PROTO_REFVALUE(p))->count & PROTO_COUNT_DATAMASK) >> 0)
#define protoSlotCount(p)    (((PROTO_REFVALUE(p))->count & PROTO_COUNT_SLOTMASK) >> 8)

// These don't normalize to a proto (from a CODE or BLOCK reference)
// So far, this is ok since all the routines that use them do a check.
#define protoSetDataCount(p,i)   ((p)->count = ((p)->count & ~PROTO_COUNT_DATAMASK) | ( (i) << 0) )
#define protoSetSlotCount(p,i)   ((p)->count = ((p)->count & ~PROTO_COUNT_SLOTMASK) | ( (i) << 8) )

#define protoSlotData(p,s)      ((Proto) (PROTO_REFVALUE(p))->data[slotIndex((s))] )
#define protoDataAtIndex(p,s)   ((Proto) (PROTO_REFVALUE(p))->data[(s)] )


struct _proto {
    uint32 header;		// Object/Malloc header and size
    uint32 count;		// Number of items in the hash
    uint32 data[1];		// This is actually 1+ entries, don't let the 1 fool you
};


typedef struct _proto * Proto;


#endif

