
//
// bublmain.c
//
// <See class's '.h' file>
//
// $Id: bublmain.c,v 1.13 2000/12/07 23:03:17 dru Exp $
//
//


#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <utime.h>
#include <ctype.h>
#include <stdlib.h>

#include "svm32.h"

#include "aqendian.h"
#include "atom.h"
#include "runtime.h"
#include "assemble.h"
#include "compiler.h"
#include "bubl2sd.h"
#include "string.h"

#include <stdio.h>
#include <sys/syscall.h>
#include <unistd.h>



int debugParse;
int debugParseLevel = -1;
int debugCodeGen;
int debugForce;


int  checkCompiled(char * fileName);
MemStream loadModuleStream(char * fileName);
int  writeModule(char * fileName, void * mem, int length);



int main (int argc, char *argv[]) 
{
    
    unsigned char * p;
    unsigned char * filename;
    int i;
    int needsCompile;

    Proto module;
    void * mem;
    int length;
    

    VM32Cpu cpu;
    VM32Cpu * pcpu;
    
    
    // Setup the memory and the GC
    

    // Setup our virtual CPU
    pcpu = &cpu;
    init(pcpu);
    reset(pcpu);


    // Setup our runtime
    bublInit(0, pcpu);

    p = UniqueString("car");
    p = UniqueString("car");
    
    for (i = 0; i < argc; i++) {
       if (argv[i][0] == '-') {
         if (strcmp(argv[i], "--trace") == 0) {
	   debugTrace = 1;
           printf("debugTrace ON\n");
         } else 
         if (strcmp(argv[i], "--dump") == 0) {
	   debugDump = 1;
           printf("debugDump ON\n");
         }
         if (strcmp(argv[i], "--force") == 0) {
	   debugForce = 1;
           printf("debugForce ON\n");
         }
         if (strcmp(argv[i], "--parse") == 0) {
	   debugParse = 1;
           printf("debugParse ON\n");
         }
         if (strcmp(argv[i], "--parseLevel") == 0) {
	     i++;
	     if (i > argc) {
		 printf("Missing argument to parseLevel\n");
		 exit(1);
	     }
	     if (!isdigit(argv[i][0])) {
		 printf("Improper argument to parseLevel\n");
		 exit(1);
	     }
	     debugParseLevel = atoi(argv[i]);
	     printf("debugParseLevel = %d ON\n", debugParseLevel);
         }
         if (strcmp(argv[i], "--codegen") == 0) {
	   debugCodeGen = 1;
           printf("debugCodeGen ON\n");
         }
       } else {
	   filename = argv[i];
       }
    }

    // Load the module
    if (argc == 1) {
	printf("Syntax: aq --parse --parseLevel <num> --codegen --trace --dump <module file>\n\n");
	printf("  parse:      Display parse tree\n");
	printf("  parseLevel: Display this level x (0=all)\n");
	printf("  codegen:    Display the generated code\n");
	printf("  trace:      Show a trace of the running code\n");
	printf("  dump:       Display the pre-compiled code module (if exists) \n");

	exit(1);
    }

    // If filename has '.aqua' suffix, then remove that
    p = strstr(filename, ".aqua");
    if (NULL != p) {
      // If the suffix is present, remove it
      *p = 0;
    }


    // Check if the module needs to be compiled to disk
    needsCompile = checkCompiled(filename);


    if (needsCompile || debugForce) {
	// Load the module
	module = compile(loadModuleStream(filename), debugParse, debugCodeGen);
	mem = bubl2sd(module, &length);
	if (writeModule(filename, mem, length)) {
	  printf("Couldn't write compiled module [%s]\n\n", filename);
	  exit(1);
	} 
    }
    

    // Prepare the module for running
    // args(filename, donotRunflag) 
    pcpu = doAquaInit(filename, 0);
    
    if (debugTrace) {
      // Turn on debug mode
      runtimeActivateTrace = 1;
    }



    // Run the code straight
    run (pcpu);
    
    return 0;
}



// checkCompiled - it checks to see if the 
int checkCompiled(char * fileName)
{
    int srcFd;
    int compiledFd;
    struct stat sb;

    unsigned int srcLen;
    unsigned int compLen;
    unsigned int srcTime;
    unsigned int compTime;

    char srcFileName[128];
    char compiledFileName[128];

    int returnVal;
    
    

    strcpy(srcFileName, fileName);
    strcat(srcFileName, ".aqua");
    
    strcpy(compiledFileName, "Compiled/");
    strcat(compiledFileName, fileName);
    strcat(compiledFileName, ".aqmod");

    
    srcFd = open(srcFileName,O_RDONLY, 0);
    if (srcFd < 0) {
        return -1;
    }
    close(srcFd);
    
    compiledFd = open(compiledFileName,O_RDONLY, 0);
    if (compiledFd < 0) {
	return -1;
    }
    close(compiledFd);
    


    stat(srcFileName, &sb);
    srcLen  = sb.st_size;
    srcTime = sb.st_mtime;
    
    stat(compiledFileName, &sb);
    compLen  = sb.st_size;
    compTime = sb.st_mtime;


    if (srcTime >= compTime) {
	// Needs a compile
	returnVal = srcTime;
    } else {
	returnVal = 0;
    }
    

    return returnVal;
    
}


int writeModule(char * fileName, void * mem, int length)
{
    int fd;
    int i;
    struct utimbuf ubuf;
    unsigned int srcTime;
    
    char realFileName[128];
    char compiledFileName[128];
    
    // Create the compiled directory if necessary
    strcpy(compiledFileName, "Compiled/");

    if ( access(compiledFileName, R_OK | X_OK | W_OK) ) {
      mkdir(compiledFileName, 0700);
      if ( access(compiledFileName, R_OK | X_OK | W_OK) ) {
        // Indicate an error
	printf("Can't created Compiled directory [%s]\n", compiledFileName);
	exit(1);
      }
    }

    // Setup the file name
    strcpy(realFileName, "Compiled/");
    strcat(realFileName, fileName);
    strcat(realFileName, ".aqmod");
    
    
    fd = open(realFileName,O_CREAT|O_WRONLY|O_TRUNC, 0777);
    if (-1 == fd) {
      // Indicate an error
      printf("Can't created compiled file [%s]\n", realFileName);
      exit(1);
    }

    i = write(fd, mem, length);

    close(fd);

    if (length != i) {
      // Indicate an error
      printf("Compiled file failed to be written to disk.");
      exit(1);
    }
    

    // Verify that we are ok time wise... 
    // Since we would only be calling write to 
    // write out a compiled version. We can use
    // the checkCompiled function to see if it
    // will work. If it won't, we'll set the time
    // on the compiled file to be 1 second more
    // 
    srcTime = checkCompiled(fileName);
    
    if (0 != srcTime) {
	ubuf.actime  = srcTime + 1;
	ubuf.modtime = srcTime + 1;
	
	utime(realFileName, &ubuf);
    }
    
    return 0;
}


	   
MemStream loadModuleStream(char * fileName)
{
    int fd;
    struct stat sb;

    char * p;
    unsigned int len;
    char realFileName[80];
    

    strcpy(realFileName, fileName);
    strcat(realFileName, ".aqua");
    
    
    
    fd = open(realFileName,O_RDONLY, 0);
    if (fd < 0) {
	printf("Couldn't open [%s], source file\n\n", realFileName);
	exit(1);
    }
//    mprintf("Fd: %d\n", (char *)fd);
    
    stat(realFileName, &sb);
    len = sb.st_size;
    
//    mprintf("Size: %d\n", (char *)len);

    p = mmap(NULL, len, PROT_READ, MAP_SHARED, fd, 0);

    return (NewMemStream(p, len));
    
}

