
//
// memstream.h
//
// This is the module that does simple memory stream work. It is very simple
// and only here to provide an easier conversion tool for working with my objc
// code.
//
//
// $Id: memstream.h,v 1.3 2000/09/22 13:05:26 dru Exp $
//
// $Log: memstream.h,v $
// Revision 1.3  2000/09/22 13:05:26  dru
// This is the post-split compiler.  The compiler was split into the node handling
// and the codegen section. The codegen section was cleaned up a lot (there is
// more to do there, I'm sure). I also got the parse/lexer errors to print
// out in a much nicer way.
//
// Revision 1.2  1999/03/29 09:20:04  dnelson
// This is a compilable, not-completely working executable version of Aqua Lang
// Added realloc to the aqmalloc
// Fixed compiler system to work
// Added an error check to the assemble system
// Reworked bublmain a little
// Built the writable memstream system
// Built the bubl2sd modules
// Added some things to the runtime
// Added the Nil object to the SimpleData format and the sd2bubl files
//
// Revision 1.1  1999/03/04 05:54:47  dnelson
// This is the first checkin of some code that is pretty stable
// This is the stack based VM for Bubl/Aqua
//
//
//
//
//


#ifndef _BUBL_MEMSTREAM_H
#define _BUBL_MEMSTREAM_H 1


struct MemStreamStruct
{
    char * start;
    uint32 size;
    uint32 cursor;
    uint32 allocated;		// memory allocated for writes
    // These will only work if you don't do any seeks! Of course, even the compiler
    // will fail this. I need to make this much more solid.
    uint32 lineNumber;
    uint32 columnNumber;
};


typedef struct MemStreamStruct * MemStream;

#ifndef EOF
#define EOF  (-1)
#endif

MemStream NewMemStream(char * start, uint32 size);

// Used by the compiler - these will have valid line number and column
// seekMS, seekMSRelative, positionMS, MSLineNumber, MScolumnNumber,
// readMSCChar


// 0 is the first byte
int seekMS(MemStream ms, uint32 where);
int seekMSRelative(MemStream ms, int where);
int getMSSize(MemStream ms);

int positionMS(MemStream ms);
int MSlineNumber(MemStream ms);
int MScolumnNumber(MemStream ms);

void getMSLineCol(MemStream ms, uint32 pos, uint32 * line, uint32 * col);
void getMSLine(MemStream ms, uint32 line, char * buff);

int readMSSizeInto(MemStream ms, uint32 size, char * into);
int readMSCInt(MemStream ms);
int readMSCChar(MemStream ms);

int addMSCSize(MemStream ms, int i);

int writeMSCSizeFrom(MemStream ms, uint32 size, char * from);
int writeMSCInt(MemStream ms, int i);
int writeMSCChar(MemStream ms, char c);




#endif
