
//
// nodes.c
//
// <see corresponding .h file>
// 
//
// $Id: nodes.c,v 1.4 2000/12/30 10:39:46 dru Exp $
//
//
//
//

#include "bubl.h"
#include "runtime.h"

// Globals


//
// Nodes codes
//

Proto createNode(Proto title)
{
    Proto tmp;
    Proto blob;

    tmp = objectNew(0);
    objectSetSlot(tmp, stringToAtom("name"), PROTO_ACTION_GET, (uint32) title);
    objectSetSlot(tmp, stringToAtom("_count"), PROTO_ACTION_GET, (uint32) objectNewInteger(-1));
    objectSetSlot(tmp, stringToAtom("_capacity"), PROTO_ACTION_GET, (uint32) objectNewInteger(8));

    blob = objectNewObjectBlob(8);
    objectSetSlot(tmp, stringToAtom("_blob"), PROTO_ACTION_GET, (uint32) blob);
    
    return tmp;
}


Proto addToNode (Proto node, Proto newItem)
{
    Proto tmp;
    int i;
    int count;
    int capacity;
    
    
    i = objectGetSlot(node, stringToAtom("_count"), &tmp);

    // A zero count means there is one

    assert(i);			// It can only add it if this is a node
    count = objectIntegerValue(tmp);
    count++;

    i = objectGetSlot(node, stringToAtom("_capacity"), &tmp);
    assert(i);			// It can only add it if this is a node
    capacity = objectIntegerValue(tmp);

    i = objectGetSlot(node, stringToAtom("_blob"), &tmp);
    assert(i);			// It can only add it if this is a node

    if (capacity == count) {
	// Time to resize
	tmp = objectBlobResize(tmp, capacity << 1);
	objectSetSlot(node, stringToAtom("_blob"), PROTO_ACTION_GET, (uint32) tmp);
	objectSetSlot(node, stringToAtom("_capacity"), PROTO_ACTION_GET, (uint32) objectNewInteger((capacity << 1)));
    }

    objectBlobAtSet(tmp, count, newItem);
    objectSetSlot(node, stringToAtom("_count"), PROTO_ACTION_GET, (uint32) objectNewInteger(count));
	
    return node;
}



Proto annotateNode(Proto node, Atom name, Proto newItem)
{
    Proto tmp;
    int i;
    
    i = objectGetSlot(node, stringToAtom("_annotate"), &tmp);

    // If it isn't there, add it
    if (0 == i) {
	tmp = objectNew(0);
	objectSetSlot(node, stringToAtom("_annotate"), PROTO_ACTION_GET, (uint32) tmp);
    }
    
    objectSetSlot(tmp, name, PROTO_ACTION_GET, (uint32) newItem);
	
    return node;
}



Proto removeFromNode(Proto node, int index)
{
    Proto tmp;
    Proto blob;
    int total, i;
    

    // Ok, loop from index +1 to end
    // Set all of the slots lower

    i = objectGetSlot(node, stringToAtom("_count"), &tmp);
    assert(i);			// It can only add it if this is a node
    
    total = objectIntegerValue(tmp);

    i = objectGetSlot(node, stringToAtom("_blob"), &blob);
    assert(i);			// It can only add it if this is a node
    
    i = index;
    
    while (1) {

	if (i == total) {
	    break;
	}
	
	tmp = objectBlobAtGet(blob, (i+1));
	objectBlobAtSet(blob, i, tmp);
	i++;
    }

    objectBlobAtSet(blob, i, nil);
    
    // Decrease the total count
    total--;
    objectSetSlot(node, stringToAtom("_count"), PROTO_ACTION_GET, (uint32) objectNewInteger(total));

    return node;
}

Proto insertNodeAt(Proto destnode, Proto node, int index)
{
    Proto tmp;
    Proto blob;
    int total, i;

    // Ok, loop from index +1 to end
    // Set all of the slots lower

    i = objectGetSlot(destnode, stringToAtom("_count"), &tmp);
    assert(i);			// It can only work if this is a node
    
    i = objectGetSlot(destnode, stringToAtom("_blob"), &blob);
    assert(i);			// It can only work if this is a node

    total = objectIntegerValue(tmp);
    i = index;

    // Catch the invalid case
    if ((i < 0) || (i > (total + 1))) {
	assert(0);
    }
    
    // Handle the simple case
    if ((total + 1) == i) {
	// Create the space
	addToNode(destnode, node);
	return destnode;
    } else {
	// Create the space
	addToNode(destnode, nil);
    }

    i = total;
    
    while (1) {

	if (i < index) {
	    break;
	}
	
	tmp = objectBlobAtGet(blob, i);
	objectBlobAtSet(blob, (i+1), tmp);
	i--;
    }

    // Finally, place the new node, buff1 should still be set right
    objectBlobAtSet(blob, (i+1), node);
    
    return destnode;
}


Proto appendNodeToNode(Proto src, Proto dest)
{
    Proto tmp;
    Proto blob;
    int i,j;
    
    
    i = objectGetSlot(src, stringToAtom("_count"), &tmp);
    assert(i);

    i = objectGetSlot(src, stringToAtom("_blob"), &blob);
    assert(i);

    
    i = objectIntegerValue(tmp);

    for (j = 0; j <= i; j++ ) {
	tmp = objectBlobAtGet(blob, j);
	addToNode(dest, tmp);
    }

    return dest;
}

uint32 nodeCount(Proto obj)
{
    Proto tmp;
    uint32 count;
    int i;
    
    i = objectGetSlot(obj, stringToAtom("_count"), &tmp);
    assert(i);

    count = objectIntegerValue(tmp);
    return count;
}

uint32 nodeGet (Proto obj, uint32 index, Proto * dest)
{
    Proto tmp;
    Proto blob;
    uint32 count;
    int i;
    
    i = objectGetSlot(obj, stringToAtom("_count"), &tmp);
    assert(i);

    count = objectIntegerValue(tmp);
    assert(count >= index);
    
    i = objectGetSlot(obj, stringToAtom("_blob"), &blob);
    assert(i);


    *dest = objectBlobAtGet(blob, index);
    
    return 1;
}

uint32 nodeSet (Proto obj, uint32 index, Proto set)
{
    Proto tmp;
    Proto blob;
    uint32 count;
    int i;
    
    i = objectGetSlot(obj, stringToAtom("_count"), &tmp);
    assert(i);

    count = objectIntegerValue(tmp);
    assert(count >= index);
    
    i = objectGetSlot(obj, stringToAtom("_blob"), &blob);
    assert(i);


    objectBlobAtSet(blob, index, set);
    
    return 1;
}


int isNode(Proto proto)
{
    Proto tmp;
    
    return (objectGetSlot(proto, stringToAtom("_count"), &tmp));
}

/*
// Compiler Node dump
void * nodeDump(Proto proto)
{
//    void * p;
    int i,j;
    Proto p1;
    Proto tmp;
    
    char buff[2048];

    // Determine the type of reference this is
    i = (uint32) proto & PROTO_REF_TYPE;
    p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

    if (proto == nil) {
	sprintf(buff, " <nil> ");
	printString(buff);
	return proto;
    }

    switch (i) {
	case PROTO_REF_INT:
	    // display the int
	    sprintf(buff, "(%ld) Integer", ((int32) proto >> 2));
	    printString(buff);
	    break;
	case PROTO_REF_PROTO:
	    // If this is a forward
	    if (p1->header & PROTO_FLAG_FORWARDER) {
		// Indicate that this is a forwarder
		sprintf(buff, "[0x%lx] Forwarder ->", ((uint32) proto & PROTO_REF_MASK));
		printString(buff);
		nodeDump((Proto)p1->count);
	    } else {
		printString("{");
		addPrintSpaces(4);
		printString("\n");
		printString("Name: ");
		objectGetSlot(proto, stringToAtom("name"), &tmp);
		sprintf(buff, "[%s] ", atomToString((Atom) tmp));
		printString(buff);
		
		printString("Count: ");
		objectGetSlot(proto, stringToAtom("_count"), &tmp);
		i = objectIntegerValue(tmp);
		sprintf(buff, "[%d]\n", i);
		printString(buff);
		
		j = objectGetSlot(proto, stringToAtom("_annotate"), &tmp);
		if (j) {
		    printString("Annotate");
		    objectDump(tmp);
		}
		    
		// enumerate through the proto showing it's contents
		j = 0;
		while (1) {
		    sprintf(buff, "\"%d\" <- ", j);
		    printString(buff);

		    // Access the data and print it
		    sprintf(buff, "%d", j);
		    objectGetSlot(proto, stringToAtom(buff), &tmp);
		    nodeDump(tmp);
		    printString("\n");
		    j++;

		    if (j > i) {
			break;
		    }
		    
		}
		addPrintSpaces(-4);
		printString("}");
	    }
	    
	    break;
	case PROTO_REF_BLOB:
	    //// Need to print the address in hex
	    sprintf(buff, "[0x%lx] Blob", ((uint32) proto & PROTO_REF_MASK));
	    printString(buff);
	    break;
	case PROTO_REF_ATOM:
	    sprintf(buff, "\"%s\"", atomToString((Atom)proto) );
	    printString(buff);
	    break;
    }

    return proto;
    
}
*/


// Remove forwards
void * removeForwards(Proto proto)
{
    int       i, j, k;
    Proto     newObj;
    Proto     blob;
    Proto     tmp;
    ProtoSlot slotPtr;

    
    // Determine the type of reference this is
    i  = PROTO_REFTYPE(proto);

    if (proto == nil) {
	return proto;
    }
    
    switch (i) {
	case PROTO_REF_PROTO:
	case PROTO_REF_BLOCK:
	case PROTO_REF_CODE:
	    // If this is a forward, flatten it
	    while (PROTO_ISFORWARDER(proto)) {
		proto = PROTO_FORWARDVALUE(proto);
	    }

	    // enumerate through the proto
	    j = 0;
	    while (1) {
		slotPtr = objectEnumerate(proto, &j);
		if (slotPtr == nil) {
		    break;
		}
		k = slotAction(slotPtr);
		switch (k) {
		    case PROTO_ACTION_GET:
		    case PROTO_ACTION_SET:
		    case PROTO_ACTION_PASSTHRU:
		    case PROTO_ACTION_CODE:
			// Access the data and print it
			newObj = removeForwards(protoSlotData(proto, slotPtr));
			objectSetSlot(proto, (Atom) slotKey(slotPtr), k, (uint32) newObj);
			break;
		}
	    }

	    
	    i = objectGetSlot(proto, stringToAtom("_count"), &tmp);
	    if (i) {
		i = objectGetSlot(proto, stringToAtom("_blob"), &blob);
		assert(i);
		j = objectIntegerValue(tmp);
		
		for (i = 0; i <= j; i++) {
		    tmp = objectBlobAtGet(blob, i);
		    newObj = removeForwards(tmp);
		    objectBlobAtSet(blob, i, newObj);
		}
	    }
	    
	    break;
    }
    return proto;
}
