
//
// runtime.c
//
// <see corresponding .h file>
// 
//
// $Id: runtime.c,v 1.20 2000/12/31 11:43:17 dru Exp $
//
// $Log: runtime.c,v $
// Revision 1.20  2000/12/31 11:43:17  dru
// This was the addition of the String objects
// This will be the first public release
//
// Revision 1.19  2000/12/30 10:39:46  dru
// This is the checkin that has working blocks!
// Plus tons of other fixes!
//
// Revision 1.18  2000/12/22 00:03:17  dru
// This is just a save point. I want to keep what we have.
// Everything compiles, the test8 compiles, but now I'm
// doing the change to the nodes.c data structure so that
// we get around the 256 slot limit.
//
// Revision 1.17  2000/12/07 23:03:17  dru
// Lots and lots of good changes.
// The system is at the point where the parse tree is much
// cleaner. It cannot generate code (again), but with this
// round of work, it should run blocks.
//
// Revision 1.16  2000/09/22 13:05:26  dru
// This is the post-split compiler.  The compiler was split into the node handling
// and the codegen section. The codegen section was cleaned up a lot (there is
// more to do there, I'm sure). I also got the parse/lexer errors to print
// out in a much nicer way.
//
// Revision 1.15  2000/08/27 19:49:28  dru
// This is the result of a major change in the code to handle
// the new reference size, and the changes in the way the
// code handles activations.
//
// Revision 1.14  2000/06/12 19:35:42  dru
// Working code that will use a 'Compiled' directory now.
//
// Revision 1.13  2000/05/01 03:55:47  dru
// This fixes many bugs I had with method invocation and parameter
// retrieval. It normalizes how frame based access is done
// in the VM.
//
// Revision 1.12  2000/04/26 01:50:04  dru
// This is a large set of changes that allowed aqua
// to do things properly. The most important of
// those was module importing. Lots of things
// were fixed (like the Lobby) and lots of intrinsics
// were added.
//
// Revision 1.11  1999/12/28 08:37:20  dru
// Show action type in objectDump
//
// Revision 1.10  1999/10/18 07:33:18  dru
// Lots and lots of changes:
// cyclic sd2bubl/bubl2sd
// fixes to doAquaInit, activateSlot, etc.
// fixes to runtime all over the place. This is a working runtime.
// This is the 'back on track' check in.
//
// Revision 1.9  1999/09/08 01:46:37  dru
// Added the hooks for dealing with output - so the debugger can
// capture the output now
//
// Revision 1.8  1999/09/07 08:48:47  dru
// Lots of changes to get the debugger working
//
// Revision 1.7  1999/08/17 15:43:01  dru
// Checking stuff in - not ready for GTK debugger
//
// Revision 1.6  1999/06/14 11:29:14  dru
//
// Added the true,false, fixed the code generation for BinarySlot protos
// to take variables properly. Fixed the codeblocks to return values
// properly.
//
// Revision 1.5  1999/05/31 08:43:25  dru
// Lots of changes in an attempt to get the 'Base' library to work.
// It still doesn't generate a working module, but it is very close.
//
// Revision 1.4  1999/04/03 08:07:00  dnelson
// This is a working version of the compiler that does integer addition
//
// Revision 1.3  1999/03/29 09:20:05  dnelson
// This is a compilable, not-completely working executable version of Aqua Lang
// Added realloc to the aqmalloc
// Fixed compiler system to work
// Added an error check to the assemble system
// Reworked bublmain a little
// Built the writable memstream system
// Built the bubl2sd modules
// Added some things to the runtime
// Added the Nil object to the SimpleData format and the sd2bubl files
//
// Revision 1.2  1999/03/28 07:08:04  dnelson
// Got the codegenerator to the point of generating code, but I haven't
// verified it yet, and I did see some errors. I changed a little bit
// of the runtime so it has an 'objectIsType' function.
//
// Revision 1.1  1999/03/04 05:54:47  dnelson
// This is the first checkin of some code that is pretty stable
// This is the stack based VM for Bubl/Aqua
//
//
//
//
//


#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#include "svm32.h"
#include "bubl.h"
#include "sd2bubl.h"
#include "runtime.h"
#include "atom.h"
#include "aqmalloc.h"
#include "assemble.h"
#include "nodes.h"
#include "compiler.h"
#include "while.h"


// INTERNAL G L O B A L S   &   H O O K S

char * bublActionTypes[] = { "Get", "Set",
			     "GetStack", "SetStack",
			     "Passthru",
			     "Code",
			     "Intrinsic", "Seven"};


int PROTO_FillFactor = 8;

Proto Lobby;
Proto DefaultBehavior;
Proto Globals;
Proto Traits;
Proto Clonable;
Proto ObjectClass;
Proto StringClass;
Proto OddBalls;

Proto IntegerClass;
Proto AtomClass;
Proto FloatClass;


// The default debugging info handler
int (*printString)(char * str);   /* Output Handler */
int (*ioOut)(char * str, int count);   /* IO Output Handler */
int (*addPrintSpaces)(int32 spaces);   /* Output Handler */
int printIndent = 0;

int runtimeActivateTrace;


static VM32Cpu * Cpu;

// Preset Atoms - we initialize these since they are used
// so often.
Atom Parent;
Atom BlockParent;
Atom Code;
Atom DebugName;


// Special Objects
Proto TrueObj;
Proto FalseObj;

Proto WhileTrueObj;
Proto WhileFalseObj;


int debugDump  = 0;
int debugTrace = 0;



// This routine will do simple string output to io
// print indents upon any newlines in the string.
int defaultIoOut(char * str, int count)
{
    write (1, str, count);
    return 0;
}


// This routine will print a string, but it will properly
// print indents upon any newlines in the string.
int defaultPrintString(char * str)
{
    char * start;
    char * end;
    
    int i;

    start = end = str;

    while (1) {
	if (*end == '\n') {

	    // If there was stuff to print before this, then print it
	    if (start != end) {
		ioOut(start, (end - start));
	    }

	    // Print spaces (the indent)
	    //// A highly inefficient indenter...
	    ioOut("\n", 1);
	    for (i = 0; i < printIndent; i++) {
		ioOut(" ", 1);
	    }
	    end++;
	    start = end;
	}

	if (*end == 0) {
	    // If there was stuff to print before this, then print it
	    if (start != end) {
		ioOut (start, (end - start));
	    }
	    break;
	}
	
	end++;
    }
    
    return 0;
}

int defaultAddPrintSpaces(int32 spaces)
{
    printIndent += spaces;

    if (printIndent < 0) {
	printIndent = 0;
	defaultPrintString("Someone botched the number of spaces to print.\n");
    }
    
    return 0;
}

int bublInit(int i, VM32Cpu * cpu) 
{
    Proto tmpObj;
    
    
    AtomInit();
    printString = defaultPrintString;
    addPrintSpaces = defaultAddPrintSpaces;
    ioOut       = defaultIoOut;
    
    Cpu = cpu;

    // Initialize some common atoms
    Parent       = stringToAtom("parent");
    BlockParent  = stringToAtom("blockParent");
    Code         = stringToAtom("_code_");
    DebugName    = stringToAtom("_debugName_");
    
    runtimeActivateTrace = 0;

    // Create the True and False objects
    tmpObj  = eval("{| true  <- {| ifTrue: a ifFalse: b <+ { a value. }. |}. false <- {| ifTrue: a ifFalse: b <+ { b value. }. |} |}", 0, 0);
    
    assembleAll(tmpObj);
    objectGetSlot(tmpObj, stringToAtom("true"),  &TrueObj);
    objectGetSlot(tmpObj, stringToAtom("false"), &FalseObj);
    Cpu->trueObj  = (uint32) TrueObj;
    Cpu->falseObj = (uint32) FalseObj;
    
    // Set up the root set, create the lobby
    Lobby = objectNew(0);

    // Create and setup DefaultBehavior
    DefaultBehavior = objectNew(0);
    objectSetSlot(Lobby, stringToAtom("DefaultBehavior"), PROTO_ACTION_GET, (uint32) DefaultBehavior);
    objectSetSlot(DefaultBehavior, stringToAtom("setSlot:with:"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicSetSlotWith);
    objectSetSlot(DefaultBehavior, stringToAtom("objectDump"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicObjectDump);
    objectSetSlot(DefaultBehavior, stringToAtom("objectShallowDump"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicObjectShallowDump);
    objectSetSlot(DefaultBehavior, stringToAtom("className"), PROTO_ACTION_GET, (uint32) stringToAtom("<DefaultBehavior>"));
    objectSetSlot(DefaultBehavior, stringToAtom("value"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicValue);

    // Build the special whileTrue:, whileFalse: slots
    objectSetSlot(DefaultBehavior, stringToAtom("whileTrue:"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicWhileTrue);
    objectSetSlot(DefaultBehavior, stringToAtom("whileFalse:"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicWhileFalse);
    WhileTrueObj = whileObj(1);
    assemble(WhileTrueObj, nil, 1);
    objectDeleteSlot(WhileTrueObj, Parent);
    
    WhileFalseObj = whileObj(0);
    assemble(WhileFalseObj, nil, 1);
    objectDeleteSlot(WhileFalseObj, Parent);
    
    
    // Setup the Traits
    Traits = objectNew(0);
    objectSetSlot(Lobby, stringToAtom("Traits"), PROTO_ACTION_GET, (uint32) Traits);
    // ..Clonable
    Clonable = objectNew(0);
    objectSetSlot(Traits, stringToAtom("Clonable"), PROTO_ACTION_GET, (uint32) Clonable);

    ObjectClass = objectNew(0);
    objectSetSlot(Clonable, stringToAtom("ObjectClass"), PROTO_ACTION_GET, (uint32) ObjectClass);
    objectSetSlot(ObjectClass, stringToAtom("className"), PROTO_ACTION_GET, (uint32) stringToAtom("Object"));
    objectSetSlot(ObjectClass, stringToAtom("clone"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicClone);

    StringClass = objectNew(0);
    objectSetSlot(Clonable, stringToAtom("StringClass"), PROTO_ACTION_GET, (uint32) ObjectClass);
    objectSetSlot(StringClass, Parent, PROTO_ACTION_GET, (uint32) ObjectClass);
    objectSetSlot(StringClass, stringToAtom("className"), PROTO_ACTION_GET, (uint32) stringToAtom("String"));
    objectSetSlot(StringClass, stringToAtom("print"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicStringPrint);
    objectSetSlot(StringClass, stringToAtom("+"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicStringAdd);
    objectSetSlot(StringClass, stringToAtom("append:"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicStringAppend);
    objectSetSlot(StringClass, stringToAtom("reverse"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicStringReverse);
    objectSetSlot(StringClass, stringToAtom("lowerCase"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicStringLowercase);
    objectSetSlot(StringClass, stringToAtom("asString"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicSelf);


    // ..OddBalls
    OddBalls = objectNew(0);
    objectSetSlot(Traits, stringToAtom("Oddballs"), PROTO_ACTION_GET, (uint32) OddBalls);
    
    IntegerClass = objectNew(0);
    objectSetSlot(OddBalls, stringToAtom("IntegerClass"), PROTO_ACTION_GET, (uint32) IntegerClass);
    objectSetSlot(IntegerClass, stringToAtom("className"), PROTO_ACTION_GET, (uint32) stringToAtom("Integer"));
    objectSetSlot(IntegerClass, stringToAtom("+"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicIntegerAdd);
    objectSetSlot(IntegerClass, stringToAtom("-"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicIntegerSubtract);
    objectSetSlot(IntegerClass, stringToAtom("*"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicIntegerMultiply);
    objectSetSlot(IntegerClass, stringToAtom("@"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicPointCreate);
    objectSetSlot(IntegerClass, stringToAtom("="), PROTO_ACTION_INTRINSIC, (uint32) intrinsicIntegerEqual);
    objectSetSlot(IntegerClass, stringToAtom(">"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicIntegerGreater);
    objectSetSlot(IntegerClass, stringToAtom("<"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicIntegerLess);
    objectSetSlot(IntegerClass, stringToAtom("print"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicIntegerPrint);
    
    AtomClass = objectNew(0);
    objectSetSlot(OddBalls, stringToAtom("AtomClass"), PROTO_ACTION_GET, (uint32) AtomClass);
    objectSetSlot(AtomClass, stringToAtom("className"), PROTO_ACTION_GET, (uint32) stringToAtom("Atom"));
    objectSetSlot(AtomClass, stringToAtom("print"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicAtomPrint);
    objectSetSlot(AtomClass, stringToAtom("size"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicAtomSize);
    objectSetSlot(AtomClass, stringToAtom("asString"), PROTO_ACTION_INTRINSIC, (uint32) intrinsicAtomAsString);

    // Setup the Global
    Globals = objectNew(0);
    objectSetSlot(Lobby, stringToAtom("Globals"), PROTO_ACTION_GET, (uint32) Globals);
    objectSetSlot(Lobby, Parent, PROTO_ACTION_GET, (uint32) Globals);

    tmpObj = objectNew(0);
    objectSetSlot(Globals, stringToAtom("Object"), PROTO_ACTION_GET, (uint32) tmpObj);
    objectSetSlot(tmpObj, Parent, PROTO_ACTION_GET, (uint32) ObjectClass);

    tmpObj = objectNew(0);
    objectSetSlot(Globals, stringToAtom("String"), PROTO_ACTION_GET, (uint32) tmpObj);
    objectSetSlot(tmpObj, Parent, PROTO_ACTION_GET, (uint32) StringClass);
    objectSetSlot(tmpObj, stringToAtom("size"), PROTO_ACTION_GET, (uint32) objectNewInteger(0));
    objectSetSlot(tmpObj, stringToAtom("capacity"), PROTO_ACTION_GET, (uint32) objectNewInteger(0));

    objectSetSlot(Globals, stringToAtom("true"),   PROTO_ACTION_GET, (uint32) TrueObj);
    objectSetSlot(Globals, stringToAtom("false"),  PROTO_ACTION_GET, (uint32) FalseObj);
    
    tmpObj = objectNew(0);
    objectSetSlot(Globals, stringToAtom("Kit"), PROTO_ACTION_GET, (uint32) tmpObj);
    objectSetSlot(tmpObj, stringToAtom("import:"),   PROTO_ACTION_INTRINSIC, (uint32) intrinsicKitImport);
    
    //// - Add the Float Routines
    FloatClass = nil;

    return 0;
}



void * objectNew(int slotNum)
{
    struct _proto * proto;
    char * p;
    
    // Malloc the space for the new object
    // the header for the object is contained as the header in the malloc chunk


    // The minimum size will be 6 slots, this will insure that we hit our fill factor properly
    slotNum = ( slotNum < 6 ? 6 : slotNum);
    
    // Ok, since the slots require 3 bits.the string literal can only take 29

    // The hash needs to know the size - the malloc's size will imply the # of slots for the max
    // table size.- That will be  (size - 4) >> 3   or (size-4)/8

    p = (char *)aqcalloc(4 + (slotNum << 3));
    p -= 4; // We want to point to the chunk header
    
    proto = (struct _proto *) p;

    // Initialize it
    proto->count = 0;

    // Now tag the header of the chunk so it is now a GC'd  'object'/proto
    proto->header |= PROTO_FLAG_GC | PROTO_FLAG_PROTO;

    // Now embed the type into the reference...
    (uint32) proto |= PROTO_REF_PROTO;

    
    return proto;
    
}

// Allocate an object that can hold 'slotNum' slots
// and 'dataNum' data slots
// Objects are basically done as 2 * the slot count
// Here we know the final size (this routine is usually
// called from sd2bubl
void * objectNewWithSize(int slotNum, int dataNum)
{

    int i;

    i = (slotNum > dataNum) ? slotNum : dataNum;

    return (objectNew(i));
}

void * objectNewInteger(int i)
{
    //// Need to check for overflows - it will happen from sd2bubl or people
    //// calling this routine directly

    // Ok, need to shift number to the left //// Hopefully this is signed
    i <<= 3;

    i |= PROTO_REF_INT;

    return (void *)i;
}



void * objectNewString(char * str)
{
    Atom p;

    p = UniqueString(str);
    
    return (void *)p;
}


void *  objectNewBlob (void * blob)
{
    Proto proto;
    char * p;

    p = blob;
    
    p -= 4; // We want to point to the chunk header
    
    proto = (struct _proto *) p;

    // Initialize it

    // Now tag the header of the chunk so it is now a GC'd  'object'/proto
    proto->header |= PROTO_FLAG_GC | PROTO_FLAG_BLOB;

    // Now embed the type into the reference...
    (uint32) proto |= PROTO_REF_BLOB;


    return ((void *) proto);
    
}

// An object blob is what stores object references for nodes, hashes, or arrays
void *  objectNewObjectBlob (int num)
{
    uchar * p;
    Proto codeBlob;

    // Build a Blob that can hold num items as capacity
    p = (uchar *) aqcalloc(num << 2);
    codeBlob = objectNewBlob(p);

    return codeBlob;
}

// What the size of a blob is
uint32 objectBlobSize (Proto obj)
{
    uint32 size;
    Proto proto;

    // Remove the type 'Blob' from the reference
    proto = PROTO_REFVALUE(obj);
    
    size = proto->header & PROTO_SIZEFIELD;
    size -= 4;			// 4 extra bytes for the malloc header

    return size;
}

// An object blob is what stores object references for nodes, hashes, or arrays
void *  objectBlobResize (Proto obj, int num)
{
    uchar * p;
    uchar * p2;
    uint32 size;
    Proto codeBlob;

    // Build a Blob that can hold num items as capacity
    p = (uchar *) aqcalloc(num << 2);

    size = objectBlobSize(obj);
    p2 = objectPointerValue(obj);

    memcpy(p, p2, size);
    
    codeBlob = objectNewBlob(p);

    return codeBlob;
}

Proto objectBlobAtGet(Proto obj, uint32 index)
{
    uint32 * p;
    
    p = (uint32 *) objectPointerValue(obj);
    return ( (Proto) p[index]);
}

void objectBlobAtSet(Proto obj, uint32 index, Proto newItem)
{
    uint32 * p;
    
    p = (uint32 *) objectPointerValue(obj);
    p[index] = (uint32)newItem;
}


//// This function will change dramatically once the GC is coded
void  objectFree(void * obj)
{
    struct _proto * proto;
    char * p;
    int i;
    
    p = (char *) obj;

    // Determine the type of reference this is
    i = (uint32) proto & PROTO_REF_TYPE;
    proto = (struct _proto *)((uint32) obj & PROTO_REF_MASK);

    switch (i) {
	case PROTO_REF_ATOM:
	case PROTO_REF_INT:
	    // Do nothing
	    break;
	case PROTO_REF_BLOB:
	case PROTO_REF_PROTO:

	    //// This will be changed dramatically once the sweep phase of the GC is coded
	    
	    // Since we are pointing within the malloc chunk header, we must restore the
	    // offset so free will work.
	    p += 4;
	    aqfree(p);
	    break;
    }
    return;

}

inline Proto objectSetAsCodeProto(Proto proto)
{
    if ( ((uint32) proto & PROTO_REF_TYPE) != PROTO_REF_PROTO) {
	assert(0);
    }

    proto = objectResolveForwarders(proto);

    // Makesure we aren't leaving any type in the pointer
    proto = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

    // Needs to be updated
    // Now tag the header of the chunk so it is now a GC'd  'object'/proto

    //// DRUDRU - must FLAG the object as code for the GC
    //proto->count |= PROTO_COUNT_TYPEISCODE;

    (uint32) proto |= PROTO_REF_CODE;
    return proto;
}

inline Proto objectSetAsBlockProto(Proto proto)
{
    if ( ((uint32) proto & PROTO_REF_TYPE) != PROTO_REF_PROTO) {
	assert(0);
    }

    proto = objectResolveForwarders(proto);

    // Makesure we aren't leaving any type in the pointer
    proto = (struct _proto *)((uint32) proto & PROTO_REF_MASK);
    (uint32) proto |= PROTO_REF_BLOCK;
    return proto;
}



inline int
objectInUse(void * obj)
{
    struct _proto * p1;
    struct _proto * p2;
    int i;
    
    p1 = (struct _proto *) ( (uint32) obj & PROTO_REF_MASK);
    p2 = (struct _proto *) ( (void *) p1 + ((uint32)p1->header & PROTO_SIZEFIELD) + 4 );

    i = p2->header & 1;

    return i;
}


// This resolves objects that are really forwarders
inline void * objectResolveForwarders(void * obj)
{
    struct _proto * p;
    
    while (1) {
	if ( PROTO_REFTYPE_REF(obj) ) {
	    return obj;
	}

	p = (struct _proto *)((uint32) obj & PROTO_REF_MASK);

	// Is the object a proto ?
	if (p->header & PROTO_FLAG_FORWARDER) {
	    obj = (struct _proto *)p->count;
	} else {
	    return (void *)( obj );
	}
	
    }
}


// Accessing type of objects
int32 objectIsType(Proto proto)
{
    int32  i;

    proto = objectResolveForwarders(proto);

    i = ((uint32) proto & PROTO_REF_TYPE); 

    return i;
}


// Accessing objects - key is a reference
void *  objectSetSlot(Proto proto, Atom key, char action, uint32 data)
{
    Proto  p1;
    ProtoSlot slotPtr;

    proto = objectResolveForwarders(proto);

    if ( PROTO_REFTYPE(proto) != PROTO_REF_PROTO ) {
        // Need to throw an exception
        // Code and Block protos should not get slots added
        assert(0);
    }

    // Lookup the key - and set a pointer to the slot

    
    if (objectLookupKey(proto, (uint32) key, &slotPtr)) {

	// If the key was found, set the new action and data..
	
	
	p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);
	
	// If it is found, set the data 
	p1->data[slotIndex(slotPtr)] = data;
	// Change the action 
	slotSetAction(slotPtr, action);
	
    } else {
        // if it isn't in there, we must add this variable to the variable array and to the hash

	// Since it will always be the slot count that overflows, we will check that
	// Check to see if hash is full and expand if necessary
	if (( (objectSlotCount(proto) * 10) / objectCapacity(proto) ) >= PROTO_FillFactor) {
	    proto = objectExpand(proto);
	    assert(!objectLookupKey(proto, (uint32) key, &slotPtr));
	}

	p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

	// Add new slot
	p1->data[protoDataCount(p1)] = data;

	// Find the empty slot for this key so we can put data into it
	assert(!objectLookupKey(proto, (uint32) key, &slotPtr));

	slotSetIndex(slotPtr, protoDataCount(p1));
	
	protoSetDataCount(p1, (protoDataCount(p1)+1)); // Now increment the data index
	slotSetKey(slotPtr, (uint32) key);
	slotSetAction(slotPtr, action);
	protoSetSlotCount(p1, (protoSlotCount(p1)+1)); // Now increment the slot count
	
    }

    return proto;
}

// This routine is for setting the data section for a proto
// It is used for sd2bubl and shouldn't be used for normal use
void *  objectSetDataAtIndex(Proto proto, uint32 index, uint32 data)
{
    uint32 i;

    proto = objectResolveForwarders(proto);

    if ( ((uint32) proto & PROTO_REF_TYPE) != PROTO_REF_PROTO) {
	//// Handle the different type
    }


    i = protoDataCount(proto);
    i = i - 1;

    // check the index
    if (index > i) {
      return nil;
    }

    protoDataAtIndex(proto, index) = data;

    return proto;
}




// This routine does an easy add of a slot that is a setter getter
// It assumes that the key is from some source, so it lacks the colon
// Example: 'x <-> 0.' 
void *  objectEasySetSlot(Proto proto, Atom key, uint32 data)
{
    char buff[64];
    
    proto = objectResolveForwarders(proto);

    if ( ((uint32) proto & PROTO_REF_TYPE) != PROTO_REF_PROTO) {
	//// Handle the different type
	assert(0);
    }

    // Overwrite existing slots and create new slot
    objectSetSlot(proto, key, PROTO_ACTION_GET, data);
    
    // Add the colon for the setter
    strcpy(buff, atomToString(key));
    strcat(buff, ":");
    
    objectSetSlotSame(proto, stringToAtom(buff), key, PROTO_ACTION_SET);

    return proto;
}



// This inserts a slot into a proto, which only affects the slot count
// This routine doesn't require the data to be passed in, just the index 
//   to the data.
// It is assumed that the data is already be present. None of the keys should already be present, otherwise it will
// overwrite the key slot. Also, it doesnt check for proto fullness
// It is assumed that those things were done (expansion and no existing slot is present)
// This is for the object passivation libraries and other routines
// internal to this module
void *  objectInsertSlot(Proto proto, uint32 key, uint32 action, uint32 index)
{
    struct _protoEntry * slotPtr;
    struct _proto * p1;
    int i;

    // Notice no check for the resolveForwarder, it shouldn't have to do that
    // I should stick in an ASSERT

    if ( ((uint32) proto & PROTO_REF_TYPE) != PROTO_REF_PROTO) {
	//// Handle the different type
        assert(0);
    }
    p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

    // Lookup the key - and set a pointer to the slot
    i = objectLookupKey(proto, key, &slotPtr);

    slotSetIndex(slotPtr, index);
    slotSetKey(slotPtr, key);
    slotSetAction(slotPtr, action);

    // If we didn't find the slot increment the slot count
    if (0 == i) {
      protoSetSlotCount(p1, (protoSlotCount(p1)+1));
    }  

    return proto;
}




// Insert a slot that references an existing data item via an existing slot
void *  objectSetSlotSame(Proto proto, Atom key, Atom existingKey, char action)
{
    struct _protoEntry * slotPtr;
    uint32   index;


    proto = objectResolveForwarders(proto);
    
    // Lookup the key - and set a pointer to the slot
    if (objectLookupKey(proto, (uint32) existingKey, &slotPtr)) {

	// Ok, get a copy of the data index for that key
	index   = slotIndex(slotPtr);

	// Check to see if hash is full and expand if necessary
	if (( (objectSlotCount(proto) * 10) / objectCapacity(proto) ) >= PROTO_FillFactor) {
	    proto = objectExpand(proto);
	}

	objectInsertSlot(proto, (uint32) key, action, index);
	
    } else {
	return nil;
    }

    return proto;
}



// Lookup the key in the hash table, if it isn't found (it will hit a nil) entry
// This routine sets the slotptr to the last slot it looked at
inline int objectLookupKey (Proto proto, uint32 key, struct _protoEntry * * slotPtr)
{
    int i,j;
    struct _proto * p1;
    

    if ( PROTO_REFTYPE_REF(proto) || (PROTO_REFTYPE(proto) == PROTO_REF_BLOB) ) {
	//// Handle the different type
        assert(0);
    }
    p1 = PROTO_REFVALUE(proto);

    //// check the types - like the key

    ////// This is a poor hash - must replace after analysis

    key &= PROTO_SLOT_KEYMASK;

    j = objectCapacity(p1);
    
    i = (key >> 11) % j;

    while (1) {
	*slotPtr = (struct _protoEntry *) &(p1->data[(j + i)]);
	if ( ( (*slotPtr)->keyActionIndex & PROTO_SLOT_KEYMASK) == key ) {
	    return 1;		// it was found, return it
	} 
	// If it is the null field, then stop
	if ( (*slotPtr)->keyActionIndex == 0 ) {
	    return 0;
	}

	i++;
	if (i == j) {
	    i = 0;
	}
    }
}


// This will expand a proto, by creating a new proto with double the capacity and then
// copying the contents of the original into the new proto. This should only be called
// from the objectSetSlot routine
void * objectExpand(struct _proto * oldProto)
{
    struct _proto * newProto;
    struct _proto * p;
    struct _protoEntry * slotPtr;
    int i;
    int oldType;
    

    if ( PROTO_REFTYPE_REF(oldProto) || (PROTO_REFTYPE(oldProto) == PROTO_REF_BLOB) ) {
	//// Handle the different type
        assert(0);
    }

    oldType = PROTO_REFTYPE(oldProto);
    // Normalize the proto to a PROTO (as opposed to CODE or BLOCK)
    oldProto = PROTO_REFVALUE(oldProto);

    // Ok, lets double the proto's size
    newProto = objectNew(objectCapacity(oldProto) << 1);

    p = (struct _proto *)((uint32)newProto & PROTO_REF_MASK);

    // First copy the instance data (it's counts
    protoSetDataCount(p, protoDataCount(oldProto));
    memcpy( p->data, oldProto->data, (objectDataCount(oldProto) << 2) );

    
    // Now lets add the contents of this hash into the other
    // hash (iterate through a proto accessing it's slots, then
    // simultaneously, add them to the new proto.
    // Since the data is already there, we use a special objectInsertSlot()
    // function that just add's the slot's and doesn't affect the data count
    // It does affect the slot count

    i = 0;
    while (1) {
	slotPtr = objectEnumerate(oldProto, &i);
	if (slotPtr == nil) {
	    break;
	}
	objectInsertSlot(newProto, slotKey(slotPtr), slotAction(slotPtr), slotIndex(slotPtr));
    }

    // Now set the original proto as a forwarder
    //// Need to make this atomic (maybe do it as a long long store?)
    oldProto->count = (uint32) p;             // have the count be the pointer to the data
    oldProto->header |= PROTO_FLAG_FORWARDER; // mark the object as a forwarder
    
    // Restore the type
    (uint32) newProto |= oldType;
    return newProto;
}

// This will clone a proto,
void * objectClone (Proto proto)
{
    Proto      newProto;
    Proto      srcProto;
    Proto      p;
    ProtoSlot  slotPtr;
    int i;
    
    proto = objectResolveForwarders(proto);
    
    // We need to get the proto's value since we access
    // the object directly and the TYPE info would offset
    // the pointer improperly
    srcProto = PROTO_REFVALUE(proto);
    i        = PROTO_REFTYPE(proto);

    switch (i) {
	case PROTO_REF_INT:
	case PROTO_REF_ATOM:
	case PROTO_REF_BLOB:
	    return proto;
	    break;
	    
	case PROTO_REF_CODE:
	    // Need to handle this
	    assert(0);
	    
	case PROTO_REF_PROTO:
	case PROTO_REF_BLOCK:
	    newProto = objectNew(objectCapacity(srcProto));
	    
	    p = (struct _proto *)((uint32)newProto & PROTO_REF_MASK);
	    
	    // First copy the instance data (it's counts
	    protoSetDataCount(p, protoDataCount(srcProto));
	    memcpy( p->data, srcProto->data, (objectDataCount(srcProto) << 2) );
	    
	    
	    // Now lets add the contents of this hash into the other
	    // hash (iterate through a proto accessing it's slots, then
	    // simultaneously, add them to the new proto.
	    // Since the data is already there, we use a special objectInsertSlot()
	    // function that just add's the slot's and doesn't affect the count
	    
	    i = 0;
	    while (1) {
		slotPtr = objectEnumerate(srcProto, &i);
		if (slotPtr == nil) {
		    break;
		}
		objectInsertSlot(newProto, slotKey(slotPtr), slotAction(slotPtr), slotIndex(slotPtr));
	    }

	    return newProto;
	    
	    break;
    }

    return nil;			// Make the compiler happy
}

    
// Enumerating through slots 
ProtoSlot objectEnumerate (Proto proto, int * currentSlot)
{
    struct _protoEntry * slotPtr;
    int i,j;

    if ( ! PROTO_REFTYPE_PROTO(proto) ) {
	//// Handle the different type
        assert(0);
    }

    
    i = objectSlotCount(proto);

    // If it is an empty proto, return
    if (i == 0) {
	return nil;
    }
    
    j = objectCapacity(proto);

    // Normalize the proto reference to an actual pointer
    proto = PROTO_REFVALUE(proto);

    // From the currentSlot number, start looking for a slot
    while (1) {

	// If we are past the end, return nil, the end-of-proto indicator
	if (*currentSlot >= j) {
	    return nil;
	} else {
	    slotPtr = (struct _protoEntry *)&( proto->data[(j  + *currentSlot)] );
	    (*currentSlot)++;	// Increment the counter
	    if ((slotPtr->keyActionIndex & 0xFFFFFFFE) != 0) {
		return slotPtr;
	    }
	}
    }
}



// Lookup the key in the hash table, if it isn't found (it will hit a nil) entry
// Return the data value for the slot
uint32 objectGetSlot(Proto proto, Atom key, Proto * result)
{
    int        i,k;
    Proto      p1;
    ProtoSlot  slotPtr;
    
    

    proto = objectResolveForwarders(proto);
    
    if ( ! PROTO_REFTYPE_PROTO(proto) ) {
	//// Handle the different type
        assert(0);
    }

    p1 = PROTO_REFVALUE(proto);

    i = objectLookupKey(p1, (uint32) key, &slotPtr);

    if (i) {
	// Check the type 
	k = slotAction(slotPtr);

	switch (k) {
	    case PROTO_ACTION_PASSTHRU:
	    case PROTO_ACTION_GET:
		*result = (Proto) protoSlotData(p1, slotPtr);
                return 1;
		break;
	    case PROTO_ACTION_GETFRAME:
		*result =  (Proto) getFrameAt(Cpu, objectIntegerValue( protoSlotData(p1, slotPtr)) );
                return 1;
		break;

		// These shouldn't be called for a getSlot
	    case PROTO_ACTION_SET:
	    case PROTO_ACTION_CODE:
	    case PROTO_ACTION_SETFRAME:
		assert(0);
		break;
		
	}
    } 

    // The slot wasn't found
    return 0;
}

// Lookup the key in the hash table, if it isn't found (it will hit a nil) entry
// If it is found, remove the slot
uint32 objectDeleteSlot (Proto proto, Atom key)
{
    int        i;
    Proto      p1;
    ProtoSlot  slotPtr;
    
    
    proto = objectResolveForwarders(proto);
    
    if ( ! PROTO_REFTYPE_PROTO(proto) ) {
	//// Handle the different type
        assert(0);
    }

    p1 = PROTO_REFVALUE(proto);

    i = objectLookupKey(p1, (uint32) key, &slotPtr);

    if (i) {
	// Determine if this is the only one pointing
	// to that index

	//// SCRATCH THAT, 
	//// 
	//// * WE'LL JUST GARBAGE COLLECT THEM LATER
	////   MAYBE WE'LL DO THAT LATER

	slotPtr->keyActionIndex = 1;
	
	// Two choices: rehash everything, kind of expensive for a delete
	//              leave a sentinel, requires a bit more work for clones, expands
	//              but is very inexpensive

	// We'll leave a sentinel, it's easy
	//protoSetSlotCount(p1, (protoSlotCount(p1)-1));


	return 1;

    } 

    // The slot wasn't found
    return 0;
}


// Return the pointer to a blob's area
void * objectPointerValue (Proto proto)
{
    struct _proto      * p1;
    char *               p;
    int  i;
    
    

    proto = objectResolveForwarders(proto);
    
    i = (uint32) proto & PROTO_REF_TYPE;

    switch (i) {

	// These shouldn't receive this call
	case PROTO_REF_INT:
	case PROTO_REF_PROTO:
	    p = nil;
	    break;
	    
	case PROTO_REF_BLOB:
	    p1 = PROTO_REFVALUE(proto);
	    p = (char *) p1;
	    // Ofset the pointer into the data
	    p += 4;
	    break;
	    
	case PROTO_REF_ATOM:
	    p = atomToString((Atom)proto);
	    break;
    }

    return p;
	
}

// Return the value of an integer reference
uint32 objectIntegerValue(Proto proto)
{
    int i;
    struct _proto      * p1;
    

    proto = objectResolveForwarders(proto);
    
    if ( PROTO_REFTYPE(proto) != PROTO_REF_INT) {
	//// Handle the different type
        assert(0);
    }

    p1 = (struct _proto *)((uint32) proto & PROTO_REF_MASK);

    
    i = (int) p1;

    // Bring the integer back
    i >>= 3;

    return (i);
}




// Return the number of data or slot spaces  in the proto 
inline int     objectCapacity(Proto proto)
{
    if ( ! PROTO_REFTYPE_PROTO(proto) ) {
	//// Handle the different type
        assert(0);
	return -1;
    }
    proto = PROTO_REFVALUE(proto);
    // TRICKY - Since we alloc with 4 + (slot * 8) 
    //          malloc really add's 4 to that
    //          So we have size = 4 + (the above)
    //          This simplifies into  header/8 = slots
    return ( ((proto->header & PROTO_SIZEFIELD) - 8) >> 3);
}


// Return the number of slots in the proto
inline int     objectSlotCount(Proto proto)
{
    if ( ! PROTO_REFTYPE_PROTO(proto) ) {
	//// Handle the different type
        assert(0);
	return -1;
    }
    return ( protoSlotCount(proto) );
}


// Return the number of slots in the proto
inline int     objectDataCount(Proto proto)
{
    if ( ! PROTO_REFTYPE_PROTO(proto) ) {
	//// Handle the different type
        assert(0);
	return -1;
    }
    return ( protoDataCount(proto) );
}



// Object dump
void * objectDump (Proto proto)
{
    int        i,j,k;
    ProtoSlot  slotPtr;
    Proto      p1;
    Proto      tmp;
    char       buff[2048];

    // Determine the type of reference this is
    i  = PROTO_REFTYPE(proto);
    p1 = PROTO_REFVALUE(proto);

    if (proto == nil) {
	sprintf(buff, " <nil> ");
	printString(buff);
	return proto;
    }
    

    switch (i) {
	case PROTO_REF_INT:
	    // display the int
	    sprintf(buff, "(%ld) Integer", objectIntegerValue(proto));
	    printString(buff);
	    break;
	case PROTO_REF_FLOAT:
	    // display the float
	    sprintf(buff, "(%ld) Float", ((int32) 0) );
	    printString(buff);
	    break;
	case PROTO_REF_PROTO:
	case PROTO_REF_BLOCK:
	case PROTO_REF_CODE:
	    if (PROTO_ISFORWARDER(proto)) {
		// Indicate that this is a forwarder
		sprintf(buff, "[0x%lx] Forwarder ->", (uint32) proto);
		printString(buff);
		objectDump(PROTO_FORWARDVALUE(proto));
	    } else {
		printString("{");
		addPrintSpaces(4);
		printString("\n");

		i = objectGetSlot(proto, stringToAtom("_count"), &tmp);
		if (i) {
		    printString("Name: ");
		    objectGetSlot(proto, stringToAtom("name"), &tmp);
		    sprintf(buff, "[%s] ", atomToString((Atom) tmp));
		    printString(buff);
		    
		    printString("Count: ");
		    objectGetSlot(proto, stringToAtom("_count"), &tmp);
		    i = objectIntegerValue(tmp);
		    sprintf(buff, "[%d]\n", i);
		    printString(buff);

		    j = objectGetSlot(proto, stringToAtom("_annotate"), &tmp);
		    if (j) {
			printString("Annotate");
			objectDump(tmp);
		    }
		    
		    
		    // enumerate through the proto showing it's contents
		    j = 0;
		    while (1) {
			if (i == -1) {
			    break;
			}
			
			sprintf(buff, "\"%d\" <- ", j);
			printString(buff);
			
			// Access the data and print it
			nodeGet(proto, j, &tmp);
			objectDump(tmp);
			printString("\n");
			j++;

			if (j > i) {
			    break;
			}
		    }
		    addPrintSpaces(-4);
		    printString("}");
		}
		else {
		    // enumerate through the proto showing it's contents
		    j = 0;
		    while (1) {
			slotPtr = objectEnumerate(proto, &j);
			if (slotPtr == nil) {
			    break;
			}
			sprintf(buff, "\"%s\" (%lx:%d) ->", atomToString((Atom)slotKey(slotPtr)), slotKey(slotPtr), j );
			printString(buff);
			k = slotAction(slotPtr);
			sprintf(buff, " (%s:%d) -> ", bublActionTypes[k], (int)slotIndex(slotPtr) );
			printString(buff);
			switch (k) {
			    case PROTO_ACTION_GET:
			    case PROTO_ACTION_SET:
			    case PROTO_ACTION_PASSTHRU:
			    case PROTO_ACTION_CODE:
				// Access the data and print it
				objectDump(protoSlotData(proto, slotPtr));
				break;
			    case PROTO_ACTION_GETFRAME:
			    case PROTO_ACTION_SETFRAME:
				// Show the For now just say stack
				tmp = protoSlotData(proto, slotPtr);
				assert(PROTO_REFTYPE(tmp) == PROTO_REF_INT);
				sprintf(buff, "FP[0x%lx]", objectIntegerValue(tmp));
				printString(buff);
				break;
			}
		    
			printString("\n");
		    }
		    addPrintSpaces(-4);
		    printString("}");
		}
	    }
	    
	    
	    break;
	case PROTO_REF_BLOB:
	    //// Need to print the address in hex
	    sprintf(buff, "[0x%lx] Blob", (uint32) proto);
	    printString(buff);
	    break;
	case PROTO_REF_CONTEXT:
	    sprintf(buff, "[0x%lx] Context", ((uint32) PROTO_REFVALUE(proto) >> 1));
	    printString(buff);
	    break;
	case PROTO_REF_ATOM:
	    sprintf(buff, "\"%s\"", atomToString((Atom)proto) );
	    printString(buff);
	    break;
    }

    return proto;
    
}

// Object dump
void * objectShallowDump(Proto proto)
{
//    void * p;
    int i,j,k;
    ProtoSlot slotPtr;
    Proto p1;
    Proto data;
    char buff[40];

    // Determine the type of reference this is
    i  = PROTO_REFTYPE(proto);
    p1 = PROTO_REFVALUE(proto);

    if (proto == nil) {
	sprintf(buff, " <nil> ");
	printString(buff);
	return proto;
    }
    

    switch (i) {
	case PROTO_REF_INT:
	    // display the int
	    sprintf(buff, "(%ld) Integer", objectIntegerValue(proto));
	    printString(buff);
	    break;
	case PROTO_REF_FLOAT:
	    // display the float
	    sprintf(buff, "(%ld) Float", ((int32) 0) );
	    printString(buff);
	    break;
	case PROTO_REF_PROTO:
	case PROTO_REF_BLOCK:
	case PROTO_REF_CODE:
	    if (PROTO_ISFORWARDER(proto)) {
		// Indicate that this is a forwarder
		sprintf(buff, "[0x%lx] Forwarder ->", (uint32) proto);
		printString(buff);
		objectShallowDump(PROTO_FORWARDVALUE(proto));
	    } else {
		printString("\n{\n");
		// enumerate through the proto showing it's contents
		j = 0;
		while (1) {
		    slotPtr = objectEnumerate(proto, &j);
		    if (slotPtr == nil) {
			break;
		    }
		    sprintf(buff, " \"%s\" ->", atomToString((Atom)slotKey(slotPtr)) );
		    printString(buff);
		    k = slotAction(slotPtr);
		    sprintf(buff, " (%s:%d) -> ", bublActionTypes[k], (int)slotIndex(slotPtr) );
		    printString(buff);
		    switch (k) {
			case PROTO_ACTION_GET:
			case PROTO_ACTION_SET:
			case PROTO_ACTION_PASSTHRU:
			case PROTO_ACTION_CODE:
			    // Access the data and print it
////DRUDRUDRU
			    data = protoSlotData(proto, slotPtr);
			    i = PROTO_REFTYPE(data);
			    if (PROTO_REF_PROTO == i) {
				sprintf(buff, "[0x%lx] Proto", (uint32) data);
				printString(buff);
			    }
			    else if (PROTO_REF_CODE == i) {
				sprintf(buff, "[0x%lx] Code", (uint32) data);
				printString(buff);
			    }
			    else {
				objectShallowDump(data);
			    }
			    break;
			case PROTO_ACTION_GETFRAME:
			case PROTO_ACTION_SETFRAME:
			    // Show the For now just say stack
			    data = protoSlotData(proto, slotPtr);
			    assert(PROTO_REFTYPE(data) == PROTO_REF_INT);
			    sprintf(buff, "FP[0x%lx]", objectIntegerValue(proto));
			    printString(buff);
			    break;
		    }
		    
		    printString("\n");
		}
		printString("}\n");
	    }
	    
	    break;
	case PROTO_REF_BLOB:
	    //// Need to print the address in hex
	    sprintf(buff, "[0x%lx] Blob", (uint32) proto);
	    printString(buff);
	    break;
	case PROTO_REF_CONTEXT:
	    sprintf(buff, "[0x%lx] Context", (uint32) proto);
	    printString(buff);
	    break;
	case PROTO_REF_ATOM:
	    sprintf(buff, "\"%s\"", atomToString((Atom)proto) );
	    printString(buff);
	    break;
    }

    return proto;
    
}


// pushHook - this is a hook into the VM's pushImmediate instruction
//            so we can clone a block in order to give it a pointer
//            to the context
//
unsigned int pushHook ( VM32Cpu * pcpu, unsigned int val) {

  Proto     codeProto;
  uint32    context;

  if (PROTO_REFTYPE(val) == PROTO_REF_BLOCK) {
     // See if we run into problems with the alignment
     assert( (((pcpu->fp << 1) & PROTO_REF_TYPE) == 0) );

     codeProto = objectClone( (Proto) val);

     // Set the parent to be the context
     context = ((pcpu->fp << 1) & PROTO_REF_MASK) | PROTO_REF_CONTEXT;
     objectSetSlot(codeProto, Parent, PROTO_ACTION_GET, (uint32) context);
     // Although this proto can never be addressed, we keep a link to it
     objectSetSlot(codeProto, BlockParent, PROTO_ACTION_GET, (uint32) val);

     // Set it to be a code type
     val = ((uint32) codeProto | PROTO_REF_CODE);
  }

  return val;

}




// activateSlot is the core routine of this language
//   it is what causes routines to be called
//   it takes it's arguments off of the virtual stack and searches
//   for the proper method
//

void activateSlot (void)
{
    
    Atom      message;
    Proto     initialTarget;
    uint32    initialFP;
    uint32    newFP;
    Proto     target;
    Proto     tmpObj;
    ProtoSlot slotPtr;
    int       ttl = 32;
    int       i,j,k;
    Proto     argCount;
    int       seenDefaultBehavior = 0;
    int       seenLobby           = 0;
    char      buff[128];
    
    
    // Ok, get the initialTarget object and the method off of the stack
    message       = (Atom)  stackAt(Cpu, 3);
    initialTarget = (Proto) stackAt(Cpu, 2);
    argCount      = (Proto) stackAt(Cpu, 1);

    // Setup the fp
    initialFP =  Cpu->fp;
    newFP     =  Cpu->fp;
    

    if (runtimeActivateTrace) {
	target = objectResolveForwarders(initialTarget);
	
	i = (uint32) target & PROTO_REF_TYPE;
	
	switch (i) {
	    case PROTO_REF_INT:
	        sprintf(buff, "target: [Int: %d] Sent: %s argCount:%d\n", (int) objectIntegerValue(target), atomToString(message), (int) objectIntegerValue(argCount));
		break;
	    case PROTO_REF_FLOAT:
	        sprintf(buff, "target: [Float: %d] Sent: %s argCount:%d\n", (int) 0, atomToString(message), (int) objectIntegerValue(argCount));
		break;
	    case PROTO_REF_ATOM:
	        sprintf(buff, "target: [Str: %s] Sent: %s argCount:%d\n", atomToString((Atom)target), atomToString(message), (int) objectIntegerValue(argCount));
		break;
	    case PROTO_REF_PROTO:
   		j = objectLookupKey(target, (uint32) DebugName, &slotPtr);
   		if (j) {
   		  objectGetSlot(target, DebugName, &tmpObj);
   		  sprintf(buff, "target: [%s] Sent: %s argCount:%d\n", atomToString((Atom)tmpObj), atomToString(message), (int) objectIntegerValue(argCount));
   		} else {
   		  sprintf(buff, "target: [0x%08lx] Sent: %s argCount:%d\n", (uint32) target, atomToString(message), (int) objectIntegerValue(argCount));
   		}
		break;
	    case PROTO_REF_CODE:
   		sprintf(buff, "target: CODE [0x%08lx] Sent: %s argCount:%d\n", (uint32) target, atomToString(message), (int) objectIntegerValue(argCount));
		break;
	    case PROTO_REF_BLOCK:
   		sprintf(buff, "target: BLOCK [0x%08lx] Sent: %s argCount:%d\n", (uint32) target, atomToString(message), (int) objectIntegerValue(argCount));
		break;
	    case PROTO_REF_CONTEXT:
   		sprintf(buff, "target: CONTEXT [0x%08lx] Sent: %s argCount:%d\n", (uint32) target, atomToString(message), (int) objectIntegerValue(argCount));
		break;
        }

	printString(buff);
    }
    
    if (nil == initialTarget) {
	sprintf(buff, "target is <nil> object. Can't send messages to nil.\n");
	printString(buff);
	exit(1);
    }
    

    if (message == stringToAtom("b")) {
	target = nil;
    }
    
    
    target = initialTarget;

    while (1) {
	
	
	target = objectResolveForwarders(target);
	
	i = (uint32) target & PROTO_REF_TYPE;
	
	switch (i) {
	    case PROTO_REF_INT:
		target = IntegerClass;
		continue;
	    case PROTO_REF_FLOAT:
		target = FloatClass;
		continue;
	    case PROTO_REF_ATOM:
		target = AtomClass;
		continue;
	    case PROTO_REF_PROTO:
	    case PROTO_REF_CODE:
		
		j = objectLookupKey(target, (uint32) message, &slotPtr);

		// If it is found
		if (j) {
		    k = slotAction(slotPtr);
		    
		    switch (k) {
			case PROTO_ACTION_GET:
			    actionGet(target, slotPtr);
			    break;
			case PROTO_ACTION_SET:
			    actionSet(target, slotPtr);
			    break;

			case PROTO_ACTION_GETFRAME:
			    actionGetStack(target, slotPtr, newFP);
			    break;
			case PROTO_ACTION_SETFRAME:
			    actionSetStack(target, slotPtr, newFP);
			    break;

			case PROTO_ACTION_PASSTHRU:
			    // Target is where we found the slot, but
			    // all of the actions on the 'self' should
			    // be on the initial target
			    actionPassthru(target, slotPtr, initialTarget);
			    break;
			case PROTO_ACTION_CODE:
			    actionCode(target, slotPtr);
			    break;
			case PROTO_ACTION_INTRINSIC:
			    actionIntrinsic(target, slotPtr);
			    break;
		    }
		    return;	// Since slot was found, we're done
		}
		// If the slot is not found
		else {
		    // Look for the parent
		    j = objectLookupKey(target, (uint32) Parent, &slotPtr);

		    // If there is a parent
		    if (j) {
			ttl--;
			if (ttl) {
			    // This needs to use the newFP
			    k = slotAction(slotPtr);
			    if (k == PROTO_ACTION_GETFRAME) {
				k = (int) protoSlotData(target, slotPtr);
				k = objectIntegerValue( (Proto) k );
				target = (Proto) getFPRelativeAt(newFP, k);
			    } else {
				objectGetSlot(target, Parent, &target);
			    }
			    
			    
			    // If this is a context reference, then we 
			    // need to use the set the newFP to that context
			    // and get the code-pointer as the new target

			    if (PROTO_REFTYPE(target) == PROTO_REF_CONTEXT) {
				newFP = ((uint32) target & PROTO_REF_MASK) >> 1;
				// Assert that this is an active context
				assert( (initialFP <= newFP) );
				target = (Proto) getFPRelativeAt(newFP, 2);
			    }
			}
			else {
			    sprintf(buff, "too many hops - method not found %s\n", atomToString(message));
			    printString(buff);
			    exit(2);
			}
			    
			continue;
		    }

		    // if there isn't a parent, try the DefaultBehavior
		    // ... if we have already done this, then a method
		    // ... really can't be found so error out
		    else {
			// If we haven't looked at the globals
			if (!seenLobby) {
			    seenLobby = 1;
			    target = Lobby;
			    continue;
			}
			if (seenDefaultBehavior) {
			    sprintf(buff, "method not found %s\n", atomToString(message));
			    printString(buff);
			    exit(3);
			}
			
			seenDefaultBehavior = 1;
			target = DefaultBehavior;
			continue;
		    }
		}
		break;
		
	} // switch
	
    } // while
    
}




// actionGet assumes that this slot was a 'GET' action. It assumes that the stack
// contained some destination object and the message activated this slot
//

inline void actionGet(Proto object, ProtoSlot slotPtr)
{
    Proto obj;
    
    // No frame is setup like in code protos, it is just a matter of getting
    // the data and returning it

    // Ok, get the data for the object
    obj = protoSlotData(object, slotPtr);
    

    // Now set that as the return value
    VMReturn(Cpu, (unsigned int) obj, 3);
    
}


// actionSet assumes that this slot was a 'SET' action. It assumes that the stack
// contained some destination object and the message activated this slot and the
// new object
//

inline void actionSet(Proto object, ProtoSlot slotPtr)
{
    // Ok, get the data for the object
    protoSlotData(object, slotPtr) = stackAt(Cpu,4);
    
    // Now return the object
    VMReturn(Cpu, (unsigned int) object, 4);
}

// actionGetStack assumes that this slot was a 'GET' action. It assumes that the stack
// contained some destination object and the message activated this slot
//

inline void actionGetStack(Proto object, ProtoSlot slotPtr, uint32 fp)
{
    uint32 obj;
    int i;
    
    // No frame is setup like in code protos, it is just a matter of getting
    // the data and returning it

    i = (int) protoSlotData(object, slotPtr);
    i = objectIntegerValue( (Proto) i );

    // Ok, get the data at that frame position
    obj = getFPRelativeAt(fp, i);
    
    // Now set that as the return value
    VMReturn(Cpu, obj, 3);
}


// actionSetStack locates the argument in the frame that has to be set and
// sets it to the argument passed in via the immediate stack. For example,
// imagine a local variable (which is stack mapped) in a code proto needs
// to get set. The trick is that the data for the set is in the immediate
// stack frame and the variable is relative to the fp
// now for the two lines of code...:)
//

inline void actionSetStack(Proto object, ProtoSlot slotPtr, uint32 fp)
{
    int i;

    i = (int) protoSlotData(object, slotPtr);
    i = objectIntegerValue( (Proto) i );

    // Ok, get the data for the frame off of the stack
    setFPRelativeAt(fp, i, stackAt(Cpu,4));
    
    // Now return the object
    VMReturn(Cpu, (unsigned int) object, 4);
}



// actionPassthru deals with activating some code for a passthru slot
//

inline void actionPassthru(Proto object, ProtoSlot slotPtr, Proto target)
{
    Proto obj;
    char * p;
    Proto codeProto;
    uint32 fp;
    
    
    // The 'object' of this message is the one that has this PASSTHRU
    // slot. However, the target is the proper parent to the code-proto.
    // So, push it onto the stack so it will be used by the code-proto's
    // stack-mapped 'parent' slot.

    // However, if it is a code-proto, then we push it's parent instead
    // and if it is a code-proto for a block, then we must go up the
    // context reference and look from there...

    // first we preserve the fp. This is because as we follow
    // context references, the next lookups of the parent (as a
    // stack relative) from within a code-proto, they will need
    // a new frame of reference
    fp = Cpu->fp;
    while (1) {
      if (PROTO_REFTYPE(target) == PROTO_REF_CODE) {
        objectGetSlot(target, Parent, &target);
      }
      if (PROTO_REFTYPE(target) == PROTO_REF_CONTEXT) {
        Cpu->fp = ((uint32)target & PROTO_REF_MASK) >> 1;
	// Get the code-proto pointed at in that context
	target  = (Proto) getFPRelativeAt(fp, 2);
	continue;
      }
      break;
    }
    // Restore the fp
    Cpu->fp = fp;
    push(Cpu, (uint32) target);

    // Get the codeProto via the data in the passThru
    codeProto = protoSlotData(object, slotPtr);

    // Now get the '_code_' from the Proto pointed at in slotPtr
    objectGetSlot(codeProto, Code, &obj);
    p = (char *) objectPointerValue(obj);
    // Push the code-proto to the stack
    push(Cpu, (uint32) codeProto);
    
    // Set the cpu to the first part of that code
    Cpu->pc = (uint32) p;

    // The code will then continue on from here, but the 'Call'
    // instruction is now complete.

    // That is, the VM will continue running bytecodes (which is 
    // what called this routine anyways), but a new activation
    // record has been created
}


// actionCode - not implemented yet.
//
inline void actionCode(Proto object, ProtoSlot slotPtr)
{
    Proto obj;
    char * p;
    Proto codeProto;
    
    
    // A marker to me that we need to implement this
    assert(0);
    // The target of the message is the code proto. However, the code
    // called in this fashion is like a 'block'. Therefore it needs to
    // have a parent. It should get the parent relative to the frame
    // pointer. The method that called this block, preserved it's 'self'
    // there.
    push(Cpu, getFrameAt(Cpu,-2));

    // Get the codeProto via the data in the passThru
    codeProto = protoSlotData(object, slotPtr);

    // Now get the '_code_' from the Proto pointed at in slotPtr
    objectGetSlot(codeProto, Code, &obj);
    p = (char *) objectPointerValue(obj);
    
    // Set the cpu to the first part of that code
    Cpu->pc = (uint32) p;

    // The code will then continue on from here, but the 'Call'
    // instruction is now complete.

    // That is, the VM will continue running bytecodes (which is 
    // what called this routine anyways), but a new activation
    // record has been created
    //// Generate an exception for now
    assert(0);
}



// actionIntrinsic calls the code via C for the argument in the slot
//

inline void actionIntrinsic(Proto object, ProtoSlot slotPtr)
{
    uint32 obj;
    void (*func)(void);
    
    
    // Ok, get the data for the object
    obj = (uint32) protoSlotData(object, slotPtr);

    func =  ( void (*)(void) ) (obj);
    

    (*func)();			// Call the function
    
    // The intrinsic takes care of the return
    
}




//////////////////////////////////////////////////////////////
//
//
// I N T R I N S I C S
//
//
//
//////////////////////////////////////////////////////////////


// Arguments: target, 'setSlot:With:', slotname/atom, new value
// Return: target
// Notes:
// If the slot doesn't exist, it goes up the parent chain
// to find a non-code proto to set the slot in
//
// This is a tricky method. If the slot exists, it will set the slot
// to a new value. This doesn't matter if the slot is a setter or a
// getter.
// Now if a slot doesn't exist, the code will look up the parent
// chain for a writable slot. Once it finds one it will create
// the slot or slots. If the name has a colon on the end, it will
// create a setter. If the name doesn't have a colon, then it will
// be just a getter
// *Note* this implies that a local variable slot cannot be created
// at runtime with this method
//
void intrinsicSetSlotWith (void)
{
    Proto      proto;
    Proto      initialProto;
    Atom       slotName;
    uint32     newData;
    uint32     initialFP;
    ProtoSlot  slotPtr;
    char       *p;
    uint32     i;
    char       buff[128];
    
    
    
    
    initialProto = proto = (Proto) stackAt(Cpu, 2);
    slotName = (Atom)   stackAt(Cpu, 4);
    newData  = (uint32) stackAt(Cpu, 5);

    // Preserve the FP
    initialFP = Cpu->fp;
    
    
    
    while (1) {

	
	proto = objectResolveForwarders(proto);

	// Try to find the slot
	if (objectLookupKey(proto, (uint32) slotName, &slotPtr)) {
	    // The key was found

	    // Set the data
	    protoSlotData(proto, slotPtr) = newData;
	    
	    // Now return the object
	    VMReturn(Cpu, (unsigned int) initialProto, 5);
	    // Restore the FP
	    Cpu->fp = initialFP;
	    return;
	}
		
	// If not there, see if this is a code proto
        if ( ((uint32) proto & PROTO_REF_TYPE) == PROTO_REF_CODE ) {
	    // Lookup parent
	    // Look for the parent
	    i = objectLookupKey(proto, (uint32) Parent, &slotPtr);
	    
	    // If there is a parent
	    if (i) {
		objectGetSlot(proto, Parent, &proto);
	        if (PROTO_REFTYPE(proto) == PROTO_REF_CONTEXT ) {
		  // We need to handle this case
		  assert(0);
		}
		continue;
	    }
	    // If no parent, this is a problem
	    else {
		//// Need to generate an exception
		sprintf(buff, "got orphaned when looking for a writable proto\n");
		printString(buff);
		// Restore the FP
		Cpu->fp = initialFP;
		exit(2);
	    }
	} else {
	    // Check if the slotName has a ':'
	    p = strchr(atomToString(slotName), ':');
	    
	    // If there is a colon
	    if (p) {
		p++;
		if (*p) {
		    //// Need to generate an exception
		    sprintf(buff, "too many colons for a setSlotWith [%s]\n", atomToString(slotName));
		    printString(buff);
		    // Restore the FP
		    Cpu->fp = initialFP;
		    exit(2);
		}
		// Create a getter
		objectSetSlot(proto, slotName, PROTO_ACTION_SET, (uint32) newData);

		// Create a setter
		strcpy(buff, atomToString(slotName));
		p = strchr(buff, ':');
		*p = 0;
		objectSetSlotSame(proto, stringToAtom(buff), slotName, PROTO_ACTION_GET);
	    } else {
		// Create a getter
		objectSetSlot(proto, slotName, PROTO_ACTION_GET, (uint32) newData);
	    }
	    
	    // Now return the object
	    VMReturn(Cpu, (unsigned int) initialProto, 5);
	    // Restore the FP
	    Cpu->fp = initialFP;
	    return;
	} // if code proto
    } // while 
    
}


// Arguments: target 'clone'
// Return: new clone
void intrinsicClone(void)
{
    Proto proto;

    proto = objectClone((Proto)stackAt(Cpu,2));
    
    VMReturn(Cpu, (unsigned int) proto, 3);
}


// Arguments: target <someMessage>
// Return: new clone
// Notes:
// Assumes some binary message, just returns the target, the self
void intrinsicSelf(void)
{
    VMReturn(Cpu, stackAt(Cpu, 2), 3);
}


// Arguments: target '+' argument
// Return: newValue
// Notes:
// Simple add function - imagine the hardware to implement this
// routine via the VM versus a simple adder circuit
void intrinsicIntegerAdd(void)
{
    int i;
    Proto newValue;

    i = objectIntegerValue( (Proto)stackAt(Cpu,2) ) +  // target
        objectIntegerValue( (Proto)stackAt(Cpu,4) );   // argument
    newValue = objectNewInteger(i);
    VMReturn(Cpu, (unsigned int) newValue, 4);
}

// Arguments: target '-' argument
// Return: newValue
// Notes:
// Simple integer subtract function
void intrinsicIntegerSubtract(void)
{
    int i;
    Proto newValue;

    i = objectIntegerValue( (Proto)stackAt(Cpu,2) ) -  // target
        objectIntegerValue( (Proto)stackAt(Cpu,4) );   // argument
    newValue = objectNewInteger(i);
    VMReturn(Cpu, (unsigned int) newValue, 4);
}

// Arguments: target '*' argument
// Return: newValue
// Notes:
// Simple integer multiply function
void intrinsicIntegerMultiply(void)
{
    int i;
    Proto newValue;

    i = objectIntegerValue( (Proto)stackAt(Cpu,2) ) *  // target
        objectIntegerValue( (Proto)stackAt(Cpu,4) );   // argument
    newValue = objectNewInteger(i);
    VMReturn(Cpu, (unsigned int) newValue, 4);
}


// Arguments: target '=' argument
// Return: true/false
// Notes:
// Simple integer equality function
void intrinsicIntegerEqual(void)
{
    int i;

    i = objectIntegerValue( (Proto)stackAt(Cpu,2) ) ==  // target
        objectIntegerValue( (Proto)stackAt(Cpu,4) );    // argument
    if (i) {
	VMReturn(Cpu, (unsigned int) TrueObj, 4);
    } else {
	VMReturn(Cpu, (unsigned int) FalseObj, 4);
    }
}

// Arguments: target '>' argument
// Return: true/false
// Notes:
// Simple integer greater-than function
void intrinsicIntegerGreater(void)
{
    int i;

    i = objectIntegerValue( (Proto)stackAt(Cpu,2) ) >  // target
        objectIntegerValue( (Proto)stackAt(Cpu,4) );    // argument
    if (i) {
	VMReturn(Cpu, (unsigned int) TrueObj, 4);
    } else {
	VMReturn(Cpu, (unsigned int) FalseObj, 4);
    }
}

// Arguments: target '<' argument
// Return: true/false
// Notes:
// Simple integer less-than function
void intrinsicIntegerLess(void)
{
    int i;

    i = objectIntegerValue( (Proto)stackAt(Cpu,2) ) ==  // target
        objectIntegerValue( (Proto)stackAt(Cpu,4) );    // argument
    if (i) {
	VMReturn(Cpu, (unsigned int) TrueObj, 4);
    } else {
	VMReturn(Cpu, (unsigned int) FalseObj, 4);
    }
}


// Arguments: target '@' argument
// Return: newValue
// Notes:
// Simple point create function 
void intrinsicPointCreate(void)
{
    Proto x, y;
    Proto point;
    Proto newPoint;
    ProtoSlot slotPtr;

    x = (Proto) stackAt(Cpu,2);
    y = (Proto) stackAt(Cpu,4);

    objectGetSlot(Globals, stringToAtom("Point"), &point);

    newPoint = objectClone(point);

    if (objectLookupKey(newPoint, (uint32) stringToAtom("x"), &slotPtr)) {
      protoSlotData(newPoint, slotPtr) = x;
    }
    if (objectLookupKey(newPoint, (uint32) stringToAtom("y"), &slotPtr)) {
      protoSlotData(newPoint, slotPtr) = y;
    }
		
    VMReturn(Cpu, (unsigned int) newPoint, 4);
}


// Arguments: target 'print'
// Return: target
// Notes:
// Prints to the default string print routine
void intrinsicIntegerPrint(void)
{
    char buff[8];

    sprintf(buff, "%ld", objectIntegerValue((Proto)stackAt(Cpu,2)) );
    printString(buff);
    VMReturn(Cpu, (unsigned int) stackAt(Cpu, 2), 3);
}


// Arguments: target 'import:' argument
// Return: target
// Notes:
// Loads in that Kit
void intrinsicKitImport(void)
{
    char buff[80];
    Atom  kitName;
    Proto kit;
    Proto root;
    Proto start;
    Proto target;
    uint32 j;
    ProtoSlot slotPtr;
    

    kitName = (Atom)stackAt(Cpu,4);
    
    // sprintf(buff, "\nImport: %s\n", atomToString(kitName) );
    // printString(buff);

    // Load the module into the system and call it's start
    // method

    // Step 1
    // Load the module into memory
    kit = intrinsicLoadModule(kitName);

    if (nil == kit) {
	sprintf(buff, "kit couldn't load [%s].Exiting.\n", atomToString(kitName) );
	printString(buff);
	exit(1);
    }

    //// Do wonderful module checking here ////

    // Ok find the root object
    objectGetSlot(kit, stringToAtom("Root"), &root);

    // Need to assemble everything...
    assembleAll(root);


    // Ok, we were called within the context of some other object.
    // We need to setup a new context with a parent local set to
    // the Root that we have here, the new tree of code that
    // we just imported. That way, when the passThru is called on 
    // Start, it will access that tree instead of from the context
    // that 'import:' was called from.


    // Resolve the forwarders 
    target = objectResolveForwarders(root);
	
    // To create a new context, we just call the actionPassthru
    // routine with a predetermined parent local. This routine
    // does the right thing
    // 
    j = objectLookupKey(target, (uint32) stringToAtom("Start"), &slotPtr);
    objectGetSlot(root, stringToAtom("Start"), &start);
    actionPassthru(target, slotPtr, target);


    //-----------------------Below this is questionable code---------------

    // OK, we are already in some code, now we need to call some
    // Usually, an intrinsic doesn't call anymore code, but we need to call
    // the '_Start' method in the module
    //
    // So, we need to make a new activation so that the return 
    // goes to the activation that called the 'import'
    //
    
    // LUCKILY, we have done nothing to the stack or processor,
    // all we really need to do is perform the actionPassthru action
    // on the data
    
    // Our context is this... 
    // activateSlot was called...
    // the slot to activate was an intrinsic...
    // we were called inside the intrinsic...
    // the program counter and arguments are on the stack...
    // 
    // we should be able to do a call in such a way that 
    // the return of this will cause the code to flow normally...
    // we need to create a new activation so that the return address
    // is the current PC?
}

Proto intrinsicLoadModule(Atom fileName)
{
    int fd;
    struct stat sb;

    int i;
    char * p;
    char buff[128];
    
    unsigned int len;
    Proto obj;
    

    sprintf(buff,"Compiled/%s.aqmod", atomToString(fileName));
    
    fd = open(buff,O_RDONLY, 0);

    if (fd < 0) {
	return nil;
    }
    
    
    stat(buff, &sb);
    len = sb.st_size;
    

    p = aqmalloc(len);
    i = read(fd, p, len);

    obj = sd2bubl(p, len);

    aqfree(p);
    
    return obj;
    
}

// Arguments: target 'value'
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a block
void intrinsicValue (void)
{
    Proto block;
    Proto obj;
    char * p;

    
    // The 'target' of this message should be a block.

    // So we need to setup the stack, it needs a frame relative
    // parent slot, which we set to nil (read the comment below)
    // then we set the secret self slot.

    // This code was borrowed heavily from the PASSTHRU action

    // The target should be a block
    block = (Proto)stackAt(Cpu,2);

    // Now get the '_code_' from the block
    objectGetSlot(block, Code, &obj);
    p = (char *) objectPointerValue(obj);

    // Push the proper parent - since this is a block, it's
    // parent isn't a GETFRAME slot, it is a CONTEXT slot.
    // Therefore, we will push 'nil' since this data isn't 
    // used
    push(Cpu, (uint32) nil);

    // Push the code-proto to the stack as the SECRET SELF
    push(Cpu, (uint32) block);
    
    // Set the cpu to the first part of that code
    Cpu->pc = (uint32) p;

    // The code will then continue on from here, but the 'Call'
    // instruction is now complete.

    // That is, the VM will continue running bytecodes (which is 
    // what called this routine anyways), but a new activation
    // record has been created
}

// Arguments: target 'whileTrue:' obj
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a WhileTrue block
void intrinsicWhileTrue (void)
{
    Proto block;
    Proto obj;
    char * p;

    
    // The 'target' of this message should be a block.

    // So we need to setup the stack, it needs a frame relative
    // parent slot, which we set to nil (read the comment below)
    // then we set the secret self slot.

    // This code was borrowed heavily from the PASSTHRU action

    // The target is the WhileTrueObj
    block = WhileTrueObj;

    // Now get the '_code_' from the block
    objectGetSlot(block, Code, &obj);
    p = (char *) objectPointerValue(obj);

    // Push the proper parent - since this is a block, it's
    // parent isn't a GETFRAME slot, it is a CONTEXT slot.
    // Therefore, we will push 'nil' since this data isn't 
    // used
    push(Cpu, (uint32) nil);

    // Push the code-proto to the stack as the SECRET SELF
    push(Cpu, (uint32) block);
    
    // Set the cpu to the first part of that code
    Cpu->pc = (uint32) p;

    // The code will then continue on from here, but the 'Call'
    // instruction is now complete.

    // That is, the VM will continue running bytecodes (which is 
    // what called this routine anyways), but a new activation
    // record has been created
}

// Arguments: target 'whileFalse:' obj
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a WhileFalse block
void intrinsicWhileFalse (void)
{
    Proto block;
    Proto obj;
    char * p;

    
    // The 'target' of this message should be a block.

    // So we need to setup the stack, it needs a frame relative
    // parent slot, which we set to nil (read the comment below)
    // then we set the secret self slot.

    // This code was borrowed heavily from the PASSTHRU action

    // The target is the WhileTrueObj
    block = WhileFalseObj;

    // Now get the '_code_' from the block
    objectGetSlot(block, Code, &obj);
    p = (char *) objectPointerValue(obj);

    // Push the proper parent - since this is a block, it's
    // parent isn't a GETFRAME slot, it is a CONTEXT slot.
    // Therefore, we will push 'nil' since this data isn't 
    // used
    push(Cpu, (uint32) nil);

    // Push the code-proto to the stack as the SECRET SELF
    push(Cpu, (uint32) block);
    
    // Set the cpu to the first part of that code
    Cpu->pc = (uint32) p;

    // The code will then continue on from here, but the 'Call'
    // instruction is now complete.

    // That is, the VM will continue running bytecodes (which is 
    // what called this routine anyways), but a new activation
    // record has been created
}


// Arguments: target 'print'
// Return: target
// Notes:
// Prints to the default string print routine
void intrinsicAtomPrint(void)
{
    printString(atomToString((Atom)stackAt(Cpu,2)));
    VMReturn(Cpu, (unsigned int) stackAt(Cpu, 2), 3);
}

// Arguments: target 'size'
// Return: integer size
// Notes:
// Returns the size of the string
void intrinsicAtomSize(void)
{
    int i;
    Proto newValue;

    i = strlen(atomToString((Atom)stackAt(Cpu,2)));
    newValue = objectNewInteger(i);
    VMReturn(Cpu, (unsigned int) newValue, 3);
}

// Arguments: target 'asString'
// Return: a new string
// Notes:
// Create a new String object out of this literal
void intrinsicAtomAsString(void)
{
    Proto target;
    int i;
    int size;
    Proto blob;
    Proto string;
    Proto newString;
    char * tp;
    char * p;
    
    target = (Proto) stackAt(Cpu,2);

    tp = atomToString((Atom)target);
    size = strlen(tp);


    objectGetSlot(Globals, stringToAtom("String"), &string);
    newString = objectClone(string);

    // Create a new blob to hold the strings
    p = (uchar *) aqcalloc(size + 1);
    blob = objectNewBlob(p);
    i = objectBlobSize(blob);
    i--; // Leave space for the null
    objectSetSlot(newString, stringToAtom("_blob"), PROTO_ACTION_GET, (uint32) blob);
    objectSetSlot(newString, stringToAtom("size"), PROTO_ACTION_GET, (uint32) objectNewInteger(size));
    objectSetSlot(newString, stringToAtom("capacity"), PROTO_ACTION_GET, (uint32) objectNewInteger(i));
    strcpy(p, tp);

    VMReturn(Cpu, (unsigned int) newString, 3);
}


// Arguments: target 'objectDump'
// Return: junk
void intrinsicObjectDump(void)
{
    objectDump((Proto)stackAt(Cpu,2));
    VMReturn(Cpu, (unsigned int) stackAt(Cpu, 2), 3);
}

// Arguments: target 'objectShallowDump'
// Return: junk
void intrinsicObjectShallowDump(void)
{
    objectShallowDump((Proto)stackAt(Cpu,2));
    VMReturn(Cpu, (unsigned int) stackAt(Cpu, 2), 3);
}


// Arguments: target 'print'
// Return: self/target
// Notes:
// Prints to the default string print routine
void intrinsicStringPrint(void)
{
    Proto target;
    int i;
    Proto blob;
    char * p;
    
    
    target = (Proto) stackAt(Cpu,2);

    i = objectGetSlot(target, stringToAtom("_blob"), &blob);
    
    if (i) {
	p = objectPointerValue(blob);
	printString(p);
    }

    VMReturn(Cpu, (unsigned int) target, 3);
}



// Arguments: target '+' literal/string
// Return: new string that is the combination
// Notes:
// Creates a new string that is the concatenation of the target and the argument
void intrinsicStringAdd(void)
{
    Proto target;
    Proto arg;
    int i;
    int size;
    int argSize;
    Proto blob;
    Proto argBlob;
    Proto string;
    Proto newString;
    char * tp;
    char * p;
    char * ap;
    
    
    target = (Proto) stackAt(Cpu,2);
    arg    = (Proto) stackAt(Cpu,4);

    // Normalize the target
    i = PROTO_REFTYPE(target);
    if (i == PROTO_REF_ATOM) {
	tp = atomToString((Atom)target);
	size = strlen(tp);
    }
    else if (i == PROTO_REF_PROTO) {
	i = objectGetSlot(target, stringToAtom("_blob"), &blob);
	if (i == 0) {
	    tp = "";
	    size = 0;
	} else {
	    tp = objectPointerValue(blob);
	    size = strlen(tp);
	}
    }

    // Normalize the argument
    i = PROTO_REFTYPE(arg);
    if (i == PROTO_REF_ATOM) {
	ap = atomToString((Atom)arg);
	argSize = strlen(ap);
    }
    else if (i == PROTO_REF_PROTO) {
	assert(objectGetSlot(arg, stringToAtom("_blob"), &argBlob));
	ap = objectPointerValue(argBlob);
	argSize = strlen(ap);
    }

    objectGetSlot(Globals, stringToAtom("String"), &string);

    newString = objectClone(string);

    // Create a new blob to hold the strings
    p = (uchar *) aqcalloc(size + argSize + 1);
    blob = objectNewBlob(p);
    i = objectBlobSize(blob);
    i--; // Leave space for the null
    objectSetSlot(newString, stringToAtom("_blob"), PROTO_ACTION_GET, (uint32) blob);
    objectSetSlot(newString, stringToAtom("size"), PROTO_ACTION_GET, (uint32) objectNewInteger(size + argSize));
    objectSetSlot(newString, stringToAtom("capacity"), PROTO_ACTION_GET, (uint32) objectNewInteger(i));
    strcpy(p, tp);
    strcat(p, ap);

    VMReturn(Cpu, (unsigned int) newString, 4);
}


// Arguments: target 'append:' literal/string
// Return: self
// Notes:
// Just appends a string on to the existing target
void intrinsicStringAppend(void)
{
    Proto target;
    Proto arg;
    int i;
    int size;
    int argSize;
    int capacity;
    Proto blob;
    Proto argBlob;
    Proto tmp;
    char * tp;
    char * p;
    char * ap;
    
    
    target = (Proto) stackAt(Cpu,2);
    arg    = (Proto) stackAt(Cpu,4);

    // Normalize the argument
    i = PROTO_REFTYPE(arg);
    if (i == PROTO_REF_ATOM) {
	ap = atomToString((Atom) arg);
	argSize = strlen(ap);
    }
    else if (i == PROTO_REF_PROTO) {
	assert(objectGetSlot(arg, stringToAtom("_blob"), &argBlob));
	ap = objectPointerValue(argBlob);
	argSize = strlen(ap);
    }

    i = objectGetSlot(target, stringToAtom("_blob"), &blob);
    
    if (i) {
	assert(objectGetSlot(target, stringToAtom("size"), &tmp));
	size = objectIntegerValue(tmp);
	assert(objectGetSlot(target, stringToAtom("capacity"), &tmp));
	capacity = objectIntegerValue(tmp);

	tp = objectPointerValue(blob);
	
	if ((size + argSize) <= (capacity + 1)) {
	    // There is room
	    objectSetSlot(target, stringToAtom("size"), PROTO_ACTION_GET, (uint32) objectNewInteger(size + argSize));
	    strcat(tp, ap);
	} else {
	    // There is no room
	    // Create a new blob to hold the strings
	    p = (uchar *) aqcalloc(size + argSize + 1);
	    blob = objectNewBlob(p);
	    i = objectBlobSize(blob);
	    i--;		// Leave space for the null
	    objectSetSlot(target, stringToAtom("_blob"), PROTO_ACTION_GET, (uint32) blob);
	    objectSetSlot(target, stringToAtom("size"), PROTO_ACTION_GET, (uint32) objectNewInteger(size + argSize));
	    objectSetSlot(target, stringToAtom("capacity"), PROTO_ACTION_GET, (uint32) objectNewInteger(i));
	    strcpy(p, tp);
	    strcat(p, ap);
	}
    } else {
	// Create the blob to hold the string
	p = (uchar *) aqcalloc(argSize+1);
	blob = objectNewBlob(p);
	i = objectBlobSize(blob);
	objectSetSlot(target, stringToAtom("_blob"), PROTO_ACTION_GET, (uint32) blob);
	objectSetSlot(target, stringToAtom("size"), PROTO_ACTION_GET, (uint32) objectNewInteger(argSize));
	objectSetSlot(target, stringToAtom("capacity"), PROTO_ACTION_GET, (uint32) objectNewInteger(i));
	strcpy(p, ap);
    }
    

    VMReturn(Cpu, (unsigned int) target, 4);
}


// Arguments: target 'reverse'
// Return: self
// Notes:
// Reverses the string in place
void intrinsicStringReverse(void)
{
    Proto target;
    int i;
    int size;
    Proto blob;
    Proto tmp;
    char * p;
    char c;
    
    target = (Proto) stackAt(Cpu,2);

    i = objectGetSlot(target, stringToAtom("_blob"), &blob);
    
    if (i) {
	i = objectGetSlot(target, stringToAtom("size"), &tmp);
	size = objectIntegerValue(tmp);
	if (size > 1) {
	    p = objectPointerValue(blob);
	    
	    // We only need to go through half
	    i = 0;
	    size--;
	    while (1) {
		// Transpose the characters
		c = p[i];
		p[i] = p[size];
		p[size] = c;
		
		i++;
		size--;
		
		if (i >= size) {
		    break;
		}
	    }
	}
    }

    VMReturn(Cpu, (unsigned int) target, 3);
}

// Arguments: target 'lowercase'
// Return: self
// Notes:
// Makes each character lower case in the string
void intrinsicStringLowercase(void)
{    Proto target;
    int i;
    Proto blob;
    char * p;
    
    target = (Proto) stackAt(Cpu,2);

    i = objectGetSlot(target, stringToAtom("_blob"), &blob);
    
    if (i) {
	p = objectPointerValue(blob);
	while (*p != 0) {
	    *p = tolower(*p);
	    p++;
	}
    }

    VMReturn(Cpu, (unsigned int) target, 3);
}

void stringAppend(Proto str, Proto obj)
{
}

//////////////////////////////////////////////////////////////
//
//
// U S E R   S U P P O R T 
//
//
//////////////////////////////////////////////////////////////




VM32Cpu * doAquaInit (char * module, int32 noRun)
{
    
    char * stack;
    int stackSize;

    Proto realRoot;
    Proto root;
    Proto start;
    
    Proto tmpObj;
    

    VM32Cpu * pcpu;
    
    

    // Setup the memory and the GC
    

    // Setup our virtual CPU
    pcpu = aqmalloc(sizeof(VM32Cpu));
    init(pcpu);

    pcpu->pushHook = pushHook;

    if (noRun) {
      pcpu->haltflag = 1; 
    }
      


    // Setup our runtime
    bublInit(0, pcpu);

    
    realRoot = loadModule(module);
    if (realRoot == nil) {
        printf("Couldn't load module named [%s]\n\n", module);
        exit(1);
    }
	
    


    // Ok find the 'Start'
    objectGetSlot(realRoot, stringToAtom("Root"), &root);
    objectGetSlot(root, stringToAtom("_Start"), &start);

    // Get the stack
    objectGetSlot(start, stringToAtom("Stack"), &tmpObj);
    stackSize = objectIntegerValue(tmpObj);
    stack = (unsigned char *) aqcalloc(stackSize);
    
    pcpu->sp = (unsigned int) (stack + stackSize - 4);
    // This stack is the type that points to where
    // the next push will go


    // Show the dump of the code (pre assembly)
    //if (debugDump) {
    if (0) {
	objectGetSlot(root, stringToAtom("Start"), &tmpObj);
	objectDump(tmpObj);
    }
    
    // Now assemble the '_Start' slot and all other Slots
    // (At least a 'Start' slot.)
    // This will make 'Start' a passthru to its assembled code proto
    assembleAll(root);


    // Run the Start code by calling 'activateSlot' on the Top with the message
    // '_Start'.
    // This 'primes the pump' with a call to activateSlot.
    // We push the context and then activateSlot, this activates the _Start
    // method. This is a 'passthru'. As such, an activate to it will just
    // setup the FP and Parent pointer (to the target) and then set the
    // PC (Program Counter) to the proper code routine.
    // Then the cpu will have to do more..


    // Push the parent
    push(pcpu, (uint32) root); 
    // Push the code-proto pointer
    push(pcpu, (uint32) root); 
    // Push the preserved FP
    push(pcpu, 0); 
    // Setup the fp
    pcpu->fp = pcpu->sp;

    push(pcpu, (uint32) stringToAtom("_Start"));
    push(pcpu, (uint32) root);
    push(pcpu, 0);		// The Argument count
    push(pcpu, 0);		// The return address


    //if (debugTrace) {
    if (0) {
      // Turn on debug mode
      runtimeActivateTrace = 1;
    }
    
    activateSlot();


    return pcpu;
}

	   
Proto loadModule(char * fileName)
{
    Proto obj;
    

    obj = intrinsicLoadModule(stringToAtom(fileName));
    

    if (debugDump) { 
      objectDump(obj);
    }

    return obj;
    
}
