
//
// runtime.h
//
// This is the description of the functions for the Aqua/Bubl runtime.
// 
//
// $Id: runtime.h,v 1.15 2000/12/31 11:43:17 dru Exp $
//
// $Log: runtime.h,v $
// Revision 1.15  2000/12/31 11:43:17  dru
// This was the addition of the String objects
// This will be the first public release
//
// Revision 1.14  2000/12/30 10:39:46  dru
// This is the checkin that has working blocks!
// Plus tons of other fixes!
//
// Revision 1.13  2000/12/22 00:03:17  dru
// This is just a save point. I want to keep what we have.
// Everything compiles, the test8 compiles, but now I'm
// doing the change to the nodes.c data structure so that
// we get around the 256 slot limit.
//
// Revision 1.12  2000/12/07 23:03:17  dru
// Lots and lots of good changes.
// The system is at the point where the parse tree is much
// cleaner. It cannot generate code (again), but with this
// round of work, it should run blocks.
//
// Revision 1.11  2000/08/27 19:49:28  dru
// This is the result of a major change in the code to handle
// the new reference size, and the changes in the way the
// code handles activations.
//
// Revision 1.10  2000/04/26 01:50:04  dru
// This is a large set of changes that allowed aqua
// to do things properly. The most important of
// those was module importing. Lots of things
// were fixed (like the Lobby) and lots of intrinsics
// were added.
//
// Revision 1.9  1999/10/18 07:33:18  dru
// Lots and lots of changes:
// cyclic sd2bubl/bubl2sd
// fixes to doAquaInit, activateSlot, etc.
// fixes to runtime all over the place. This is a working runtime.
// This is the 'back on track' check in.
//
// Revision 1.8  1999/09/08 01:46:37  dru
// Added the hooks for dealing with output - so the debugger can
// capture the output now
//
// Revision 1.7  1999/09/07 08:48:47  dru
// Lots of changes to get the debugger working
//
// Revision 1.6  1999/08/17 15:43:01  dru
// Checking stuff in - not ready for GTK debugger
//
// Revision 1.5  1999/06/14 11:29:14  dru
//
// Added the true,false, fixed the code generation for BinarySlot protos
// to take variables properly. Fixed the codeblocks to return values
// properly.
//
// Revision 1.4  1999/05/31 08:43:25  dru
// Lots of changes in an attempt to get the 'Base' library to work.
// It still doesn't generate a working module, but it is very close.
//
// Revision 1.3  1999/04/03 08:07:00  dnelson
// This is a working version of the compiler that does integer addition
//
// Revision 1.2  1999/03/28 07:08:05  dnelson
// Got the codegenerator to the point of generating code, but I haven't
// verified it yet, and I did see some errors. I changed a little bit
// of the runtime so it has an 'objectIsType' function.
//
// Revision 1.1  1999/03/04 05:54:48  dnelson
// This is the first checkin of some code that is pretty stable
// This is the stack based VM for Bubl/Aqua
//
//
//
//
//


#include "svm32.h"
#include "bubl.h"
#include "atom.h"

#ifndef H_BUBLRUNTIME
#define H_BUBLRUNTIME



#define nil   (void *)0
#define True  1
#define False 0




#define PROTO_ACTION_GET          0
#define PROTO_ACTION_SET          1
#define PROTO_ACTION_GETFRAME     2
#define PROTO_ACTION_SETFRAME     3
#define PROTO_ACTION_PASSTHRU     4
#define PROTO_ACTION_CODE         5
#define PROTO_ACTION_INTRINSIC    6




// G L O B A L S   &   H O O K S

// Fill factor for the proto hashes
extern int PROTO_FillFactor;

// Fill factor for the proto hashes
extern Proto Lobby;
extern Proto DefaultBehavior;
extern Proto Globals;
extern Proto Traits;
extern Proto Clonable;
extern Proto ObjectClass;
extern Proto OddBalls;

extern Proto IntegerClass;
extern Proto AtomClass;

// Special Objects
extern Proto TrueObj;
extern Proto FalseObj;


extern int debugParse;
extern int debugParseLevel;
extern int debugDump;
extern int debugTrace;


// The default debugging info handler
extern int (*printString)(char * str);        /* Output handler */
extern int (*ioOut)(char * str, int count);   /* IO Output Handler */
extern int (*addPrintSpaces)(int32 spaces);   /* Output Handler */

int runtimeActivateTrace;


extern int runtimeActivateTrace;



int     bublInit(int i, VM32Cpu * cpu);


void *  objectNew(int slotNum);
void *  objectNewWithSize(int slotNum, int dataNum);
void *  objectNewInteger(int i);
void *  objectNewString(char * str);
void *  objectNewBlob(void * blob);
// An object blob is what stores object references for nodes, hashes, or arrays
void *  objectNewObjectBlob(int num);
// What the size of a blob is
uint32 objectBlobSize (Proto obj);
void *  objectBlobResize (Proto obj, int num);
Proto   objectBlobAtGet(Proto obj, uint32 index);
void    objectBlobAtSet(Proto obj, uint32 index, Proto newItem);

void    objectFree(void * obj);


inline Proto   objectSetAsCodeProto(Proto proto);
inline Proto   objectSetAsBlockProto(Proto proto);
inline int     isObjectSetAsCodeProto(Proto proto);

inline void  * objectResolveForwarders(void * obj);


// Accessing objects -> these follow forwarders before accessing
// object
int32 objectIsType(Proto proto);
//void *  objectForKeyString(void * proto, char * str);
//void *  objectForKey(void * proto, char * key);
//void *  objectSlotAt(void * proto, int slotNum);
void *  objectSetSlot(Proto proto, Atom key, char action, uint32 data);

void *  objectSetDataAtIndex(Proto proto, uint32 index, uint32 data);

void *  objectEasySetSlot(Proto proto, Atom key, uint32 data);

void *  objectInsertSlot(Proto proto, uint32 key, uint32 action, uint32 index);

// Set another key/action to point to the same data as an existingKey
void *  objectSetSlotSame(Proto proto, Atom key, Atom existingKey, char action);

// Hash type lookup
inline int objectLookupKey(Proto proto, uint32 key, struct _protoEntry * * slotPtr);

// Accessing data from objects
//int     objectInteger(void * proto);
//void *  objectData(void * proto);

void * objectExpand(Proto proto);

// Enumerating through slots
ProtoSlot objectEnumerate(Proto proto, int * currentSlot);

// This gets the data from a slot, it returns true if it
// finds the key
//// Whether this function will exist is a good question
uint32 objectGetSlot(Proto proto, Atom key, Proto * result);

// This removes a slot from a proto 
// it returns true if it finds the key
uint32 objectDeleteSlot(Proto proto, Atom key);


// This returns a pointer to the data - this only works on blobs
void *  objectPointerValue(Proto proto);

uint32    objectIntegerValue(Proto proto);


inline int objectCapacity(Proto proto);
inline int objectSlotCount   (Proto proto);
inline int objectDataCount   (Proto proto);

// Dump the object to stdio
void *  objectDump(Proto proto);
void *  objectShallowDump(Proto proto);


// The magic function
// This function expects the proto, selector, and arguments on the vm stack
void activateSlot(void);



/////////////////////////////////////////////////////
//
// A C T I O N   H A N D L E R S
//
//


// Action Handlers - inlined because only activateSlot should call them
inline void actionGet(Proto object, ProtoSlot slotPtr);

// actionSet assumes that this slot was a 'SET' action. It assumes that the stack
// contained some destination object and the message activated this slot and the
// new object
//
inline void actionSet(Proto object, ProtoSlot slotPtr);


// actionGetStack assumes that this slot was a 'GET' action. It assumes that the stack
// contained some destination object and the message activated this slot
//
inline void actionGetStack(Proto object, ProtoSlot slotPtr, uint32 fp);


// actionSetStack locates the argument in the frame that has to be set and
// sets it to the argument passed in via the immediate stack. For example,
// imagine a local variable (which is stack mapped) in a code proto needs
// to get set. The trick is that the data for the set is in the immediate
// stack frame and the variable is relative to the fp
// now for the two lines of code...:)
//
inline void actionSetStack(Proto object, ProtoSlot slotPtr, uint32 fp);


// actionPassthru deals with activating some code for a passthru slot
//
// This will push the parent if 'useParamAsParent' is true otherwise
// it will retrieve the parent off of the stack. This is the target
// of the message that had the passThru slot.
inline void actionPassthru(Proto object, ProtoSlot slotPtr, Proto parent);


// actionCode - not implemented yet.
//
inline void actionCode(Proto object, ProtoSlot slotPtr);

// actionIntrinsic calls the code via C for the argument in the slot
//
inline void actionIntrinsic(Proto object, ProtoSlot slotPtr);


/////////////////////////////////////////////////////
//
// I N T R I N S I C S
//
//


void intrinsicSetSlotWith(void);

// Arguments: target 'clone'
// Return: new clone
void intrinsicClone(void);

// Arguments: target <someMessage>
// Return: new clone
// Notes:
// Assumes some binary message, just returns the target, the self
void intrinsicSelf(void);

// Arguments: target '+' argument
// Return: newValue
// Notes:
// Simple add function - imagine the hardware to iplement this versus
// a simple adder circuit
void intrinsicIntegerAdd(void);


// Arguments: target '-' argument
// Return: newValue
// Notes:
// Simple integer subtract function
void intrinsicIntegerSubtract(void);

// Arguments: target '*' argument
// Return: newValue
// Notes:
// Simple integer multiply function
void intrinsicIntegerMultiply(void);

// Arguments: target '=' argument
// Return: true/false
// Notes:
// Simple integer equality function
void intrinsicIntegerEqual(void);

// Arguments: target '>' argument
// Return: true/false
// Notes:
// Simple integer greater-than function
void intrinsicIntegerGreater(void);

// Arguments: target '<' argument
// Return: true/false
// Notes:
// Simple integer less-than function
void intrinsicIntegerLess(void);

// Arguments: target '@' argument
// Return: newValue
// Notes:
// Simple point create function 
void intrinsicPointCreate(void);

// Arguments: target 'print'
// Return: target
// Notes:
// Prints to the default string print routine
void intrinsicIntegerPrint(void);

// Arguments: target 'import:' argument
// Return: target
// Notes:
// Loads in that Kit
void intrinsicKitImport(void);

// A support method for KitImport
Proto intrinsicLoadModule(Atom fileName);

// Arguments: target 'value'
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a block
void intrinsicValue(void);

// Arguments: target 'whileTrue:' obj
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a WhileTrue block
void intrinsicWhileTrue (void);

// Arguments: target 'whileFalse:' obj
// Return: NONE - the block does the return
// Notes:
// This changes the thread of control to a WhileFalse block
void intrinsicWhileFalse (void);


// Arguments: target 'print'
// Return: target
// Notes:
// Prints to the default string print routine
void intrinsicAtomPrint(void);

// Arguments: target 'size'
// Return: integer size
// Notes:
// Returns the size of the string
void intrinsicAtomSize(void);

// Arguments: target 'asString'
// Return: a new string
// Notes:
// Create a new String object out of this literal
void intrinsicAtomAsString(void);


// Arguments: target 'objectDump'
// Return: junk
void intrinsicObjectDump(void);

// Arguments: target 'objectShallowDump'
// Return: junk
void intrinsicObjectShallowDump(void);


// Arguments: target 'print'
// Return: self/target
// Notes:
// Prints to the default string print routine
void intrinsicStringPrint(void);

// Arguments: target '+' literal/string
// Return: new string that is the combination
// Notes:
// Creates a new string that is the concatenation of the target and the argument
void intrinsicStringAdd(void);

// Arguments: target 'append:' literal/string
// Return: self
// Notes:
// Just appends a string on to the existing target
void intrinsicStringAppend(void);

// Arguments: target 'reverse'
// Return: self
// Notes:
// Reverses the string in place
void intrinsicStringReverse(void);

// Arguments: target 'lowercase'
// Return: self
// Notes:
// Makes each character lower case in the string
void intrinsicStringLowercase(void);


// U S E R  S U P P O R T functions


VM32Cpu * doAquaInit (char * module, int32 noRun);

Proto loadModule(char * fileName);


#endif
